# Copyright European Space Agency, 2013

"""
This module provides core classes to handle georeferenced images.
Most classes are base classes which are inherited from in other modules
of this package.
"""

from __future__ import division, print_function, absolute_import

from six.moves import range, map
from six import add_metaclass

import logging
import time
from abc import ABCMeta, abstractproperty, abstractmethod
from collections import namedtuple
import copy

import numpy as np
import numpy.ma as ma

import astropy.units as u
from astropy import constants as const
from astropy.coordinates.angles import Angle

from auromat.coordinates.transform import j2000ToLatLon, rotatePole, geoToMLatMLT,\
    geodetic2Ecef, smToLatLon, mltToSmLon, j2000ToMLatMLT
from auromat.coordinates.geodesic import wgs84A, wgs84B, containsOrCrossesPole
import auromat.utils
from auromat.coordinates import geodesic
from auromat.coordinates.geodesic import Location
from auromat.coordinates.intersection import sphereLineIntersection,\
    ellipsoidLineIntersection
from auromat.util.image import loadImage
from auromat.utils import outline, convexHull, polygonCentroid, extend
from auromat.util.decorators import lazy_property, inherit_docs
import warnings

Size = namedtuple('Size', ['width','height'])
PixelScales = namedtuple('PixelScales', ['width', 'height', 'diagonal'])
PixelScale = namedtuple('PixelScale', ['mean', 'median', 'min', 'max'])

class BoundingBox(object):
    """
    Describes a geographical bounding box that can span across
    the discontinuity.
    """
    def __init__(self, latSouth, lonWest, latNorth, lonEast):
        """
        
        :param number latSouth: in degrees [-90,90]
        :param number lonWest: in degrees [-180,180]
        :param number latNorth: in degrees [-90,90]
        :param number lonEast: in degrees [-180,180]
        """
        assert -180 <= lonWest <= 180, 'Longitude: ' + str(lonWest)
        assert -180 <= lonEast <= 180, 'Longitude: ' + str(lonEast)
        assert -90 <= latSouth <= 90, 'Latitude: ' + str(latSouth)
        assert -90 <= latNorth <= 90, 'Latitude: ' + str(latNorth)
        
        self._latSouth = latSouth
        self._lonWest = lonWest
        self._latNorth = latNorth
        self._lonEast = lonEast
        
    @property
    def latSouth(self):
        return self._latSouth
    
    @property
    def lonWest(self):
        return self._lonWest
    
    @property
    def latNorth(self):
        return self._latNorth
    
    @property
    def lonEast(self):
        return self._lonEast
    
    @property
    def topLeft(self):
        """
        North west corner of the bounding box.
        
        :rtype: auromat.coordinates.geodesic.Location
        """
        return Location(self.latNorth, self.lonWest)
    
    @property
    def bottomLeft(self):
        """
        South west corner of the bounding box.
        
        :rtype: auromat.coordinates.geodesic.Location
        """
        return Location(self.latSouth, self.lonWest)
    
    @property
    def topRight(self):
        """
        North east corner of the bounding box.
        
        :rtype: auromat.coordinates.geodesic.Location
        """
        return Location(self.latNorth, self.lonEast)
    
    @property
    def bottomRight(self):
        """
        South east corner of the bounding box.
        
        :rtype: auromat.coordinates.geodesic.Location
        """
        return Location(self.latSouth, self.lonEast)
    
    @lazy_property
    def _minSphericalRectangle(self):
        """
        based on https://stackoverflow.com/a/13503064
        
        useful as stereographic projection parameters for drawing
        
        Note that this currently only works for bounding boxes spanning
        less than 180 degrees in longitude. For greater longitude ranges
        the returned center is still correct but the size is wrong.
        
        :rtype: tuple (center, size) with size a named tuple (width, height) with values in km
                and center a named Location tuple
        """
        if self.containsPole:
            if self.latNorth == 90: # north pole
                center = Location(90, 0)
                width = geodesic.distance(center, Location(self.latSouth, 0))*2
            else: # south pole
                center = Location(-90, 0)
                width = geodesic.distance(center, Location(self.latNorth, 0))*2              
            height = width            
        else:
            lonWest = self.lonWest
            lonEast = self.lonEast
            if lonWest > lonEast:
                lonEast += 360
            lonc = (lonWest+lonEast)/2
            lonc = Angle(lonc * u.deg).wrap_at(180 * u.deg).degree
            
            if lonEast - lonWest > 180:
                warnings.warn('The bounding box spans more than 180deg in longitude. '
                              'The returned size of the minimum spherical rectangle will be incorrect.')
            
            # TODO only works when longitude range is less than 180 deg
            # -> this is because the geodesic distance is not oriented and always
            #    returns the shortest distance
            width = geodesic.distance(self.bottomLeft, self.bottomRight)
            width2 =  geodesic.distance(self.topLeft, self.topRight)
            if width2 > width:
                # southern hemisphere
                width = width2
                bottomCenter = geodesic.intermediate(self.bottomLeft, self.bottomRight, 0.5)
                topDataCenter = Location(self.latNorth, lonc)
                height = geodesic.distance(topDataCenter, bottomCenter)
                center = geodesic.intermediate(topDataCenter, bottomCenter, 0.5)            
            else:
                # northern hemisphere
                topCenter = geodesic.intermediate(self.topLeft, self.topRight, 0.5)
                bottomDataCenter = Location(self.latSouth, lonc)
                height = geodesic.distance(bottomDataCenter, topCenter)
                center = geodesic.intermediate(bottomDataCenter, topCenter, 0.5)
        
        return center, Size(width/1000, height/1000)
    
    @lazy_property
    def center(self):
        """
        Center of the minimum spherical rectangle that fits the
        bounding box.
        
        :rtype: auromat.coordinates.geodesic.Location
        """
        center, _ = self._minSphericalRectangle
        return center
    
    @lazy_property
    def size(self):
        """
        Width and height in km of the minimum spherical rectangle that fits the
        bounding box.
        
        Note that the size is only correct if the bounding box spans
        less than 180 degrees in longitude.
        
        :rtype: Size
        """
        _, size = self._minSphericalRectangle
        return size
    
    @property
    def containsDiscontinuity(self):
        """
        Return whether the bounding box contains the 180 degree discontinuity.
        
        :rtype: bool
        """
        return self.lonWest > self.lonEast or self.containsPole
    
    @property
    def containsPole(self):
        """
        Return whether the bounding box contains the north and/or south pole.
        
        :rtype: bool
        """
        return self.lonWest == -180 and self.lonEast == 180 and \
               (self.latNorth == 90 or self.latSouth == -90)
    
    @staticmethod
    def minimumBoundingBox(latLons):
        """
        Return the smallest bounding box that contains all given
        coordinates.
        
        :param latLons: iterable of (lat,lon) pairs
        :rtype: BoundingBox
        """
        bbs = [BoundingBox(lat, lon, lat, lon) for [lat,lon] in latLons]        
        minBB = BoundingBox.mergedBoundingBoxes(bbs)               
        return minBB
    
    @staticmethod
    def mergedBoundingBoxes(boundingBoxes):
        """
        Return the smallest bounding box that contains all given
        bounding boxes.
        
        :param boundingBoxes: iterable of BoundingBox instances
        :rtype: BoundingBox
        """
        boundingBoxes = list(boundingBoxes)
        lats = [bb.latSouth for bb in boundingBoxes] + [bb.latNorth for bb in boundingBoxes]       
        lons = [(bb.lonWest,bb.lonEast) for bb in boundingBoxes]
        
        latSouth = np.min(lats)
        latNorth = np.max(lats)
        lonWest, lonEast = BoundingBox._minimumBoundingBoxLons(lons)
        
        return BoundingBox(latSouth, lonWest, latNorth, lonEast)
    
    @staticmethod
    def _minimumBoundingBoxLons(lons):       
        lons = np.asarray(lons)
        xs = np.sort(lons.ravel())
        assert len(xs) % 2 == 0
        xs = np.concatenate((xs,[xs[0]+360]))
        
        lonsUnwrapped = np.rad2deg(np.unwrap(np.deg2rad(lons)))
    
        coversInterval = np.zeros(len(xs)-1, dtype=bool)
        for i in range(1, len(xs)):
            for bb in lonsUnwrapped:
                if bb[0] <= xs[i-1] and bb[1] >= xs[i]:
                    coversInterval[i-1] = True
                    break
        
        intervalLengths = xs[1:] - xs[:-1]
        gapLengths = ma.masked_array(intervalLengths, coversInterval)
        biggestGapIdx = np.argmax(gapLengths) # ignoring other possible maximums
        
        lonWest, lonEast = xs[biggestGapIdx+1], xs[biggestGapIdx]
        lonWest = Angle(lonWest * u.deg).wrap_at(180 * u.deg).degree
        lonEast = Angle(lonEast * u.deg).wrap_at(180 * u.deg).degree
                
        return lonWest, lonEast
    
    def __eq__(self, obj):
        return isinstance(obj, BoundingBox) and \
           self.latNorth == obj.latNorth and self.latSouth == obj.latSouth and \
           self.lonWest == obj.lonWest and self.lonEast == obj.lonEast
           
    def _ne__(self, obj):
        return not self == obj
    
    def __repr__(self):
        return 'BoundingBox(latSouth={0}, lonWest={1}, latNorth={2}, lonEast={3})'.format(
                       self.latSouth, self.lonWest, self.latNorth, self.lonEast)

MappingProperties = namedtuple('MappingProperties', 
                               'altitude cameraPosGCRS boundingBox photoTime '
                               'centroid cameraFootpoint identifier')

@add_metaclass(ABCMeta)
class BaseMapping(object):    
    """
    Base class for all mapping objects.
    Describes a georeferenced image for a given altitude.
    
    The following assertions always hold:
    If a pixel in img is defined, then it is guaranteed that the corresponding
    center and corner coordinates are defined, as well as the elevation.
    If a pixel in img is NOT defined, then the corresponding center coordinate and
    elevation are also not defined; one of its corner coordinates may be defined
    in case a neighbouring pixel is defined, otherwise it is not defined.
    
    More formally:
    
    - If lats[y,x] is not masked, then lons[y,x] is not masked, and vice versa.
    - If latsCenter[y,x] is not masked, then lonsCenter[y,x] is not masked, and vice versa.
    - If lats[y,x] is not masked, then at least one of latsCenter[y,x], latsCenter[y,x-1],
      latsCenter[y-1,x-1] or latsCenter[y-1,x] is not masked.
    - If latsCenter[y,x] is not masked, then lats[y,x], lats[y+1,x],
      lats[y+1,x+1] and lats[y,x+1] are not masked.
    - If img[y,x] is not masked, then latsCenter[y,x] is not masked, and vice versa.
    - If elevation[y,x] is not masked, then latsCenter[y,x] is not masked, and vice versa.
    """
    
    def __init__(self, altitude, cameraPosGCRS, photoTime, identifier, metadata=None):
        """
            
        :param number altitude: the altitude in km onto which the image is mapped                    
        :param array-like cameraPosGCRS: [x,y,z] camera position in GCRS coordinates, in km
        :param datetime.datetime photoTime: date the photo was taken
        :param str identifier: a string which uniquely names this mapping (e.g. ISS030-E-84614);
                               it must be usable for parts of a filename
        :param dict metadata: a flat dictionary of informational metadata that is used when
                              exporting a mapping
        """
        assert altitude >= 0
        cameraPosGCRS = np.asarray(cameraPosGCRS)
        assert cameraPosGCRS.shape == (3,)

        self._altitude = altitude
        self._cameraPosGCRS = cameraPosGCRS
        self._photoTime = photoTime
        self._identifier = identifier
        self._metadata = metadata
        
        # cached values of derived attributes
        self._centroid = None
        self._outlines = None
        self._boundingBox = None
        self._pixelScales = None
        self._mlatmlt = None
        self._mlatmltCenter = None
        
    @property
    def properties(self):
        """
        Return information about the mapping as a named tuple.
        
        :rtype: MappingProperties
        """
        return MappingProperties(identifier=self.identifier,
                                 altitude=self.altitude,
                                 cameraPosGCRS=self.cameraPosGCRS,
                                 boundingBox=self.boundingBox,
                                 photoTime=self.photoTime,
                                 centroid=self.centroid,
                                 cameraFootpoint=self.cameraFootpoint)
    
    def checkGuarantees(self):
        """
        Checks if the guarantees defined for coordinate and data arrays are fulfilled.
        See the docstring of this class.
        
        Note: This method is meant only for testing purposes.
        """
        lats, lons = self.lats, self.lons
        latsCenter, lonsCenter = self.latsCenter, self.lonsCenter
        mlat, mlt = self.mLatMlt
        mlatCenter, mltCenter = self.mLatMltCenter
        img = self.img
        elevation = self.elevation
        
        # we use masked arrays, so there must be no NaN's in the unmasked parts
        assert not np.any(np.isnan(lats))
        assert not np.any(np.isnan(latsCenter))
        assert not np.any(np.isnan(mlat))
        assert not np.any(np.isnan(elevation))     
        
        # If lats[y,x] is not nan, then lons[y,x] is not nan, and vice versa.
        assert np.all(np.logical_xor(ma.getmaskarray(lats), ~ma.getmaskarray(lons)))
        
        # If latsCenter[y,x] is not nan, then lonsCenter[y,x] is not nan, and vice versa.
        assert np.all(np.logical_xor(ma.getmaskarray(latsCenter), ~ma.getmaskarray(lonsCenter)))
        
        # If lats[y,x] is not nan, then at least one of latsCenter[y,x], latsCenter[y,x-1],
        # latsCenter[y-1,x-1] or latsCenter[y-1,x] is not nan.
        latsCenterNotNanPadded = np.zeros((latsCenter.shape[0]+2, latsCenter.shape[1]+2), bool)
        latsCenterNotNanPadded[1:-1,1:-1] = ~ma.getmaskarray(latsCenter)
        assert np.all(np.logical_or.reduce((ma.getmaskarray(lats),
                                            latsCenterNotNanPadded[1:,1:],
                                            latsCenterNotNanPadded[1:,:-1],
                                            latsCenterNotNanPadded[:-1,:-1],
                                            latsCenterNotNanPadded[:-1,1:]
                                            )))
        
        # If latsCenter[y,x] is not nan, then lats[y,x], lats[y+1,x],
        # lats[y+1,x+1] and lats[y,x+1] are not nan.
        latsNotNan = ~ma.getmaskarray(lats)
        assert np.all(np.logical_or(ma.getmaskarray(latsCenter),
                                    np.logical_and.reduce((latsNotNan[:-1,:-1],
                                                           latsNotNan[1:,:-1],
                                                           latsNotNan[1:,1:],
                                                           latsNotNan[:-1,1:]
                                                           ))))
        
        # If img[y,x] is not nan, then latsCenter[y,x] is not nan, and vice versa.
        latsCenterNotNan = ~ma.getmaskarray(latsCenter)
        imgNan = ma.getmaskarray(img)
        for d in range(img.shape[2]):
            assert np.all(np.logical_xor(imgNan[:,:,d], latsCenterNotNan))
                
        # If elevation[y,x] is not nan, then latsCenter[y,x] is not nan, and vice versa.          
        assert np.all(np.logical_xor(ma.getmaskarray(elevation), latsCenterNotNan))
        
        # If mlatCenter[y,x] is not nan, then latsCenter[y,x] is not nan, and vice versa.
        assert np.all(np.logical_xor(ma.getmaskarray(mlatCenter), latsCenterNotNan))
        
        # If mltCenter[y,x] is not nan, then latsCenter[y,x] is not nan, and vice versa.
        assert np.all(np.logical_xor(ma.getmaskarray(mltCenter), latsCenterNotNan))
        
        # If mlat[y,x] is not nan, then lats[y,x] is not nan, and vice versa.
        assert np.all(np.logical_xor(ma.getmaskarray(mlat), latsNotNan))
        
        # If mlt[y,x] is not nan, then lats[y,x] is not nan, and vice versa.
        assert np.all(np.logical_xor(ma.getmaskarray(mlt), latsNotNan))
        
    @property
    def altitude(self):
        """
        Mapping altitude in km.
        """
        return self._altitude
    
    @property
    def cameraPosGCRS(self):
        return self._cameraPosGCRS
    
    @lazy_property
    def cameraFootpoint(self):
        """
        The camera footpoint in geodetic coordinates.
        
        :rtype: auromat.coordinates.geodesic.Location
        """
        camFootLat, camFootLon = j2000ToLatLon([self.cameraPosGCRS], self.photoTime)
        return Location(camFootLat[0], camFootLon[0])
    
    @property
    def photoTime(self):
        """
        :rtype: datetime.datetime
        """
        return self._photoTime
    
    @property
    def identifier(self):
        """
        :rtype: str
        """
        return self._identifier
    
    @property
    def metadata(self):
        """
        :rtype: dict
        """
        if self._metadata is None:
            return {}
        else:
            return self._metadata
        
    @abstractproperty
    def lats(self):
        """
        A masked array of shape (img.shape[0]+1, img.shape[1]+1)
        containing the latitude of every pixel corner.
        """
    
    @abstractproperty
    def lons(self):
        """
        A masked array of shape shape (img.shape[0]+1, img.shape[1]+1)
        containing the longitude of every pixel corner.
        """
    
    @abstractproperty
    def latsCenter(self):
        """
        A masked array of shape shape (img.shape[0], img.shape[1])
        containing the latitude of every pixel center.
        """

    @abstractproperty
    def lonsCenter(self):
        """
        A masked array of shape (img.shape[0], img.shape[1])
        containing the longitude of every pixel center.
        """
        
    @property
    def isPlateCarree(self):
        """
        Return whether the latitude/longitude arrays describe a plate carree
        projection (regular grid).
        """
        return isPlateCarree(self.lats, self.lons)
    
    def checkPlateCarree(self):
        """
        Raises an error when the latitude/longitude arrays do not describe a plate carree
        projection (regular grid).
        """
        return checkPlateCarree(self.lats, self.lons)
        
    @property
    def mLatMlt(self):
        """
        Tuple of masked arrays of the geomagnetic latitude/magnetic local time of every pixel corner.
        
        :rtype: tuple(mlat, mlt) with array shape (img.shape[0]+1, img.shape[1]+1).
        """
        if self._mlatmlt is None:
            self._mlatmlt = self._mLatMlt(self.lats, self.lons)
        return self._mlatmlt
    
    @property
    def mLatMltCenter(self):
        """
        Tuple of masked arrays of the geomagnetic latitude/magnetic local time of every pixel center.
        
        :rtype: tuple(mlat, mlt) with shape (img.shape[0], img.shape[1]).
        """
        if self._mlatmltCenter is None:
            self._mlatmltCenter = self._mLatMlt(self.latsCenter, self.lonsCenter)
        return self._mlatmltCenter
    
    def _mLatMlt(self, lats, lons):
        mask = ma.getmaskarray(lats)
        lats, lons = np.deg2rad(lats.data), np.deg2rad(lons.data)
        geoX, geoY, geoZ = geodetic2Ecef(lats, lons, self.altitude, wgs84A, wgs84B)
        geo = np.dstack((geoX,geoY,geoZ))
        mlat, mlt = geoToMLatMLT(geo.reshape(-1,3), self.photoTime)
        mlat = mlat.reshape(geo.shape[0], geo.shape[1])
        mlt = mlt.reshape(geo.shape[0], geo.shape[1])
        mlat = ma.masked_array(mlat, mask)
        mlt = ma.masked_array(mlt, mask)
        return mlat, mlt
    
    @abstractproperty
    def img(self):
        """
        Masked array of shape (h,w,n) and type (u)int.
        """
        
    @abstractproperty
    def img_unmasked(self):
        """
        Like img but as a normal numpy array (not a masked array).
        
        The purpose of this property (compared to img.data) is to implement
        more efficient ways to access the image data, in the case that
        the coordinates (lat, lon, etc.) are not needed. These would be
        calculated to properly mask the image.
        """
    
    @abstractproperty
    def elevation(self):
        """
        Elevation in degrees for each pixel center, that is,
        a masked array of shape (img.shape[0], img.shape[1]).
        """
    
    @abstractproperty
    def rgb(self):
        """
        A representation of img as a masked (h,w,3) uint8 RGB image with a range [0,255].
        
        This is a convenience property for easy display of the image data.
        
        This array shall never be modified directly, only through
        the underlying img.
        
        Note to implementers: If this property is cached then care must be
        taken when a mapping gets masked. For example, when implementing
        :meth:`createMasked` then the cache may be invalidated.
        """
    
    @abstractproperty
    def rgb_unmasked(self):
        """
        Like rgb but as a normal numpy array (not a masked array).
        
        The purpose of this property (compared to rgb.data) is to implement
        more efficient ways to access the image data, in the case that
        the coordinates (lat, lon, etc.) are not needed. These would be
        calculated to properly mask the image.
        """
        
    @abstractmethod
    def createResampled(self, lats, lons, latsCenter, lonsCenter, elevation, img):
        """
        Returns a new mapping object with the given values.
        Each subclass can decide what the most appropriate class for
        the resampled data is. E.g. ThemisMapping uses itself, while
        AstrometryMapping uses GenericMapping. This is important as
        only the original class knows how to interpret the img data
        (if it's not standard RGB in the case of THEMIS).
        See :func:`auromat.resample.resample` for an application of this method.
        
        :param elevation: can be None
        """

    @abstractmethod    
    def createMasked(self, centerMask):
        """
        Return a copy of this mapping with the given mask applied
        to img, latsCenter, lonsCenter, and elevation.
        Implementing classes must override this method and handle
        the lats and lons attributes and possible others.
        See :meth:`maskedByElevation` for an application of this method.
        
        :param centerMask: the mask is not copied and should not be
                           used after anymore after calling this method
        """
        # mask is copied, data is not      
        _latsCenter = ma.masked_array(self.latsCenter.data, centerMask)
        _lonsCenter = ma.masked_array(self.lonsCenter.data, centerMask)
        _img = ma.masked_array(self.img.data, centerMask)
        _elevation = ma.masked_array(self.elevation.data, centerMask)

        class MaskedMapping(object):                
            @property
            def latsCenter(self):
                return _latsCenter
            
            @property
            def lonsCenter(self):
                return _lonsCenter
            
            @property
            def img(self):
                return _img
            
            @property
            def elevation(self):
                return _elevation
            
        m = copy.copy(self)
        extend(m, MaskedMapping)
        return m
            
    @property
    def outline(self):
        """
        The complete outline of this mapping. 
        Note that the outline can be concave.
        """
        full, _ = self._fullAndConvexOutlines
        return full
    
    @property
    def outlineConvexHull(self):
        """
        The convex hull of the regular outline.
        """
        _, convex = self._fullAndConvexOutlines
        return convex
    
    @property
    def _fullAndConvexOutlines(self):
        if self._outlines is None:
            mask = ~ma.getmaskarray(self.lats)            
            outl = outline(mask)
            
            # full outline
            outlLats = self.lats[outl[:,1], outl[:,0]]
            outlLons = self.lons[outl[:,1], outl[:,0]]
            outlLatLon = np.transpose([outlLats, outlLons])
            
            # convex hull
            outlConvex = convexHull(outl)
            outlConvexLats = self.lats[outlConvex[:,1], outlConvex[:,0]]
            outlConvexLons = self.lons[outlConvex[:,1], outlConvex[:,0]]
            outlConvexLatLon = np.transpose([outlConvexLats, outlConvexLons])
            
            self._outlines = outlLatLon, outlConvexLatLon    
            
        return self._outlines        
    
    @property   
    def boundingBox(self):
        """
        In case containsPole is True, this bounding box is degenerate!
        It will span the full longitude range in this case.
        """
        if self._boundingBox is None:
            outlLats = self.outline[:,0]
            outlLons = self.outline[:,1]
            
            latMin, latMax = np.min(outlLats), np.max(outlLats)
            lonMin, lonMax = np.min(outlLons), np.max(outlLons)
            
            # To make pole checking faster we reduce the outline.
            # Removing points from the full outline could produce a 
            # self-intersecting polygon, therefore the convex hull of
            # the full outline is used such that this cannot happen.
            pointCount = len(self.outlineConvexHull)
            sampleCount = min(pointCount, 50)
            indices = np.round(np.linspace(0, pointCount-1, sampleCount)).astype(np.int)           
            reducedOutline = self.outlineConvexHull[indices]
            
            if containsOrCrossesPole(reducedOutline):
                lonWest = -180
                lonEast = 180
                # find out which pole!
                if latMax < 0: # south pole
                    latSouth = -90
                    latNorth = latMax
                else: # north pole
                    latNorth = 90
                    latSouth = latMin
            else:
                # we assume that mappings are smaller than 180deg longitudes
                # and use this assumption to determine if the mapping contains
                # the discontinuity
                if lonMax-lonMin > 180:
                    # contains discontinuity
                    westLons = outlLons>0
                    eastLons = ~westLons
                    lonWest = np.min(outlLons[westLons])
                    lonEast = np.max(outlLons[eastLons])
                else:
                    lonWest = lonMin
                    lonEast = lonMax
                latNorth, latSouth = latMax, latMin
            
            bb = BoundingBox(latSouth, lonWest, latNorth, lonEast)
            self._boundingBox = bb
        
        return self._boundingBox
    
    @property
    def containsDiscontinuity(self):
        """
        Return whether the bounding box contains the discontinuity.
        """
        return self.boundingBox.containsDiscontinuity
        
    @property
    def containsPole(self):
        """
        Return whether the bounding box contains a pole.
        """
        return self.boundingBox.containsPole
    
    @property
    def centroid(self):
        """The centroid of the mapping based on the plate-carree projection.
        
        :rtype: auromat.coordinates.geodesic.Location
        """        

        if self._centroid is None:
            if self.containsPole:
                # TODO rotate away from pole, see resample module
                raise NotImplementedError
            elif self.containsDiscontinuity:
                lats = self.outline[:,0]
                lons = self.outline[:,1]
                lons = Angle((lons + 180) * u.deg).wrap_at(180 * u.deg).degree
                outline = np.transpose([lats, lons])
                
                lat,lon = polygonCentroid(outline)
                
                centroidLon = Angle((lon + 180) * u.deg).wrap_at(180 * u.deg).degree
                self._centroid = Location(lat, centroidLon)
            else:
                lat,lon = polygonCentroid(self.outline)
                self._centroid = Location(lat, lon)
        
        return self._centroid
    
    @property
    def arcSecPerPx(self):
        """
        Min, max, median, and mean angular sizes of pixels/polygons determined for the width, height, and
        diagonal of 1000 polygons.
        
        :rtype: PixelScales
        """
        if self._pixelScales is None:            
            # create polygons without colors 
            # (copied from auromat.draw.createPolygonsAndColors)
            latLonDeg = ma.dstack((self.lats, self.lons))
            verts = ma.concatenate((
                        latLonDeg[0:-1, 0:-1],
                        latLonDeg[0:-1, 1:  ],
                        latLonDeg[1:  , 1:  ],
                        latLonDeg[1:  , 0:-1],
                        ), axis=2)
            verts = verts.reshape((latLonDeg.shape[0]-1)*(latLonDeg.shape[1]-1), 4, 2)
            hasNans = ma.getmaskarray(verts).any(axis=-1).any(axis=-1)
            verts = verts[~hasNans]
            
            # calculate angular sizes for 1000 polygons using the polygon width
            # (just 1000 because geographiclib is quite slow due to no array support)
            polyCount = verts.shape[0]
            sampleCount = min(polyCount, 1000)
            indices = np.round(np.linspace(0, polyCount-1, sampleCount)).astype(np.int)
            
            widthSizesDeg = [geodesic.angularDistance(Location(poly[0][0], poly[0][1]), 
                                                      Location(poly[1][0], poly[1][1]))
                             for poly in verts[indices]]

            heightSizesDeg = [geodesic.angularDistance(Location(poly[1][0], poly[1][1]), 
                                                       Location(poly[2][0], poly[2][1]))
                              for poly in verts[indices]]
            
            diagonalSizesDeg = [geodesic.angularDistance(Location(poly[0][0], poly[0][1]), 
                                                         Location(poly[2][0], poly[2][1]))
                                for poly in verts[indices]]
            
            # reduce to some useful values
            scales = []
            for sizesDeg in [widthSizesDeg, heightSizesDeg, diagonalSizesDeg]:
                meanDegPerPx = np.mean(sizesDeg)
                medianDegPerPx = np.median(sizesDeg)
                minDegPerPx = min(sizesDeg)
                maxDegPerPx = max(sizesDeg)
                
                meanArcSecPerPx = (meanDegPerPx * u.deg).to(u.arcsec).value
                medianArcSecPerPx = (medianDegPerPx * u.deg).to(u.arcsec).value
                minArcSecPerPx = (minDegPerPx * u.deg).to(u.arcsec).value
                maxArcSecPerPx = (maxDegPerPx * u.deg).to(u.arcsec).value
                
                scales.append(PixelScale(meanArcSecPerPx, medianArcSecPerPx, minArcSecPerPx, maxArcSecPerPx))
            
            self._pixelScales = PixelScales(width=scales[0], height=scales[1], diagonal=scales[2])
            
        return self._pixelScales
        
    def maskedByElevation(self, minElevation=10):
        """
        Return a new mapping with data masked below the given minimum elevation.
        A previously applied mask is ignored.
        
        Note that the new mapping reuses the existing arrays and just exchanges
        the masks.
        
        :param int minElevation: 0 to 90, in degrees
        :rtype: BaseMapping
        """
        assert self.elevation is not None
        centerMask = (self.elevation < minElevation).filled(True)
        if np.all(centerMask):
            raise ValueError('minElevation=' + str(minElevation) + ' would mask all pixels!')
        
        newMapping = self.createMasked(centerMask)
        newMapping.setDirty()
                
        return newMapping
        
    def maskedByPolygon(self, polygon):
        """
        Returns a copy of this mapping where the image is masked using the
        given polygon. Only those pixels are retained where all of its corners
        are inside the polygon. A previously applied mask is ignored.
        
        Note that the new mapping reuses the existing arrays and just exchanges
        the masks.
        
        .. warning:: If the mapping or the polygon contains the discontinuity and/or poles then
                 this method tries to handle it in a best-effort approach.
                 It may happen that this approach fails where the resulting mapping could
                 contain invalid data.
        
        :param array-like polygon: ordered points of an unclosed polygon in [lat,lon] order
        :rtype: BaseMapping
        """
        polygon = np.asarray(polygon)
        latLonGrid = np.dstack((self.lats.data, self.lons.data))
        latLonGridFlat = latLonGrid.reshape(-1, 2)

        polyBoundingBox = BoundingBox.minimumBoundingBox(polygon)
        polyContainsPole = containsOrCrossesPole(polygon)   
        if self.containsDiscontinuity or polyBoundingBox.containsDiscontinuity:
            angle = 180
            polygon = polygon.copy()
            for arr in [latLonGridFlat, polygon]:
                arr[:,1] = Angle((arr[:,1] + angle) * u.deg).wrap_at(angle * u.deg).degree
                
        elif self.containsPole or polyContainsPole:
            angle = 90
            xaxis = [1,0,0]
            for arr in [latLonGridFlat, polygon]:
                arr[:,0], arr[:,1] = rotatePole(np.deg2rad(arr[:,0]), 
                                                np.deg2rad(arr[:,1]),
                                                self.altitude, 
                                                angle=angle, axis=xaxis)
              
        isInside = auromat.utils.pointsInsidePolygon(latLonGridFlat, polygon)
        isInside = isInside.reshape(self.lats.shape)
        mask = ~isInside | self.lats.mask

        if np.all(mask):
            raise ValueError('The given mask would mask all pixels!')
        
        # we mask every pixel which misses at least one of its four corner coordinates        
        centerMask = np.logical_or.reduce((mask[:-1,:-1], mask[1:,:-1], mask[:-1,1:], mask[1:,1:]))
        
        newMapping = self.createMasked(centerMask)
        newMapping.setDirty()
                
        return newMapping
    
    def setDirty(self):
        """
        Sets all values of cached properties to None so that they are
        recalculated the next time they are accessed.
        """
        self._outlines = None
        self._boundingBox = None
        self._centroid = None
        self._pixelScales = None
        self._mlatmlt = None
        self._mlatmltCenter = None

def checkPlateCarree(lats, lons):
    """
    Checks whether the given 2D coordinate arrays describe a
    plate carree projection, that is, if the latitudes (longitudes)
    are evenly spaced and monotonically decreasing (increasing).
    
    :param ndarray lats: 2D latitude array
    :param ndarray lons: 2D longitude array
    :raise ValueError: when the projection is not plate carree 
    """
    if ma.isMaskedArray(lats):
        lats, lons = lats.data, lons.data
    if np.any(np.isnan(lats)):
        raise ValueError('coordinates contains NaNs')
    lons = np.unwrap(np.deg2rad(lons))
    # x=lon must be monotonically increasing
    # y=lat must be monotonically decreasing
    if lons[0,-1]-lons[0,0] <= 0:
        raise ValueError('longitudes are not monotonically increasing')
    if lats[0,0]-lats[-1,0] <= 0:
        raise ValueError('latitudes are not monotonically decreasing')
    # lat and lon must be evenly spaced
    eps = 1e-4
    deltaLon = lons[0,1:] - lons[0,:-1]
    isLonRegular = np.max(deltaLon) - np.min(deltaLon) < eps
    if not isLonRegular:
        raise ValueError('longitudes are not evenly spaced')
    deltaLat = lats[:-1,0] - lats[1:,0]
    isLatRegular = np.max(deltaLat) - np.min(deltaLat) < eps
    if not isLatRegular:
        raise ValueError('latitudes are not evenly spaced')

def isPlateCarree(lats, lons):
    """
    Return whether the given 2D coordinate arrays describe a
    plate carree projection, that is, if the latitudes (longitudes)
    are evenly spaced and monotonically decreasing (increasing).
    
    :param ndarray lats: 2D latitude array
    :param ndarray lons: 2D longitude array
    :rtype: bool
    """
    try:
        checkPlateCarree(lats, lons)
    except:
        return False
    else:
        return True
      
class DefaultRGBMixin(object):
    """
    Mixin for use with :class:`BaseMapping`. 
    Interprets an 8 bit (uint8) or 16 bit (uint16) 3-channel image as RGB image
    with the channels in RGB order. 
    """
     
    @property
    def rgb(self):
        return ma.masked_array(self.rgb_unmasked, mask=self.img.mask)
        
    @property
    def rgb_unmasked(self):
        if self.img_unmasked.dtype == np.uint8:
            img = self.img_unmasked
        elif self.img_unmasked.dtype == np.uint16:
            img = self.img_unmasked.copy()
            img *= 255/65535
            img = img.astype(np.uint8)
        else:
            raise NotImplementedError
        
        if img.shape[2] == 3:
            return img
        elif img.shape[2] == 1:
            return np.repeat(img, 3, 2)
        else:
            raise NotImplementedError('Unknown img format')

class ArrayImageMixin(DefaultRGBMixin):
    """
    Mixin for use with :class:`BaseMapping`. 
    Interprets an 8 bit (uint8) or 16 bit (uint16) 3-channel array as RGB image
    with the channels in RGB order. 
    """
    
    def __init__(self, img):
        assert img.ndim == 3
        assert img.dtype in [np.uint8, np.uint16]
        if not ma.isMaskedArray(img):
            img = ma.masked_array(img)
        self._img = img
    
    @property
    def img(self):
        return self._img
    
    @property
    def img_unmasked(self):
        return self._img.data
            
class FileImageMixin(DefaultRGBMixin):
    """
    Mixin for use with :class:`BaseMapping`. 
    Loads an 8 or 16 bit image file and converts it to an RGB image
    in case it is grayscale.
    """
    
    # TODO don't convert grayscale to RGB, only for rgb property later on
    #      -> relevant for MIRACLE
    
    def __init__(self, imagePath):
        self._imagePath = imagePath
        self._img = None
        self._img_unmasked = None
    
    @property
    def img(self):
        if self._img is None:
            self._img = ma.masked_array(self.img_unmasked)
        return self._img
    
    @property
    def img_unmasked(self):
        if self._img_unmasked is None:
            img = loadImage(self._imagePath)
            self._img_unmasked = img
        return self._img_unmasked
    
    @property
    def imagePath(self):
        return self._imagePath

def _doSanitize(lats, lons, latsCenter, lonsCenter, img, elevation, afterMasking=False):
    """
    Sanitizes the given masked data arrays such that _checkGuarantees() returns True.
    This method works directly on the masks of the given arrays.
    
    :param afterMasking: If True, skips steps which are not needed when sanitizing
                         an already sanitized mapping that has been masked.
    """
    t0 = time.time()
    
    centerMaskShared = isinstance(latsCenter.mask, np.ndarray) and \
                       latsCenter.mask is lonsCenter.mask is elevation.mask
    cornerMaskShared = isinstance(lats.mask, np.ndarray) and lats.mask is lons.mask
    
    # first, apply img mask to center coordinates
    imgMask = ma.getmaskarray(img)[:,:,0]
    latsCenter[imgMask] = ma.masked
    if not centerMaskShared:
        lonsCenter[imgMask] = ma.masked
    
    # now, make sure lats,lons,latsCenter,lonsCenter are consistent
    latsCenterNanPadded = np.ones((latsCenter.shape[0]+2, latsCenter.shape[1]+2), bool)
    latsCenterNanPadded[1:-1,1:-1] = ma.getmaskarray(latsCenter)

    allNeighboursMissing = np.logical_and.reduce((latsCenterNanPadded[1:,1:],
                                                  latsCenterNanPadded[1:,:-1],
                                                  latsCenterNanPadded[:-1,:-1],
                                                  latsCenterNanPadded[:-1,1:]))

    lats[allNeighboursMissing] = ma.masked
    if not cornerMaskShared:
        lons[allNeighboursMissing] = ma.masked
    
    if not afterMasking:
        latsNan = ma.getmaskarray(lats)
        anyCornerMissing = np.logical_or.reduce((latsNan[:-1,:-1],
                                                 latsNan[1:,:-1],
                                                 latsNan[1:,1:],
                                                 latsNan[:-1,1:]))
    
        latsCenter[anyCornerMissing] = ma.masked
        if not centerMaskShared:
            lonsCenter[anyCornerMissing] = ma.masked
        
        # and again in case there are new corners without neighbors
        latsCenterNanPadded = np.ones((latsCenter.shape[0]+2, latsCenter.shape[1]+2), bool)
        latsCenterNanPadded[1:-1,1:-1] = ma.getmaskarray(latsCenter)
    
        allNeighboursMissing = np.logical_and.reduce((latsCenterNanPadded[1:,1:],
                                                      latsCenterNanPadded[1:,:-1],
                                                      latsCenterNanPadded[:-1,:-1],
                                                      latsCenterNanPadded[:-1,1:]))
    
        lats[allNeighboursMissing] = ma.masked
        if not cornerMaskShared:
            lons[allNeighboursMissing] = ma.masked
    
    # finally, apply coords mask back to img and elevation
    img[ma.getmaskarray(latsCenter)] = ma.masked
    if not centerMaskShared:        
        elevation[ma.getmaskarray(img)[:,:,0]] = ma.masked
    
    print('sanitize:', time.time()-t0, 's')

def sanitize_data(cls):
    """
    A decorator for classes inheriting from :class:`BaseMapping`. It lazily sanitizes data such that
    :meth:`BaseMapping.checkGuarantees` is satisfied. This decorator assumes the most general case. 
    If a more optimized sanitization technique can be implemented
    for a specific mapping class, then this should be done and this decorator not be used.
    """
    @inherit_docs
    class SanitizedMapping(cls):
                
        @property
        def lats(self):
            return self._sanitized(lambda s: s.lats)
        
        @property
        def lons(self):
            return self._sanitized(lambda s: s.lons)
        
        @property
        def latsCenter(self):
            return self._sanitized(lambda s: s.latsCenter)
        
        @property
        def lonsCenter(self):
            return self._sanitized(lambda s: s.lonsCenter)
        
        @property
        def img(self):
            return self._sanitized(lambda s: s.img)
        
        @property
        def elevation(self):
            return self._sanitized(lambda s: s.elevation)
            
        def createMasked(self, centerMask):
            # mask is copied, data is not
            latslonsmask = ma.getmaskarray(self.lats).copy()
            _lats = ma.masked_array(self.lats.data, latslonsmask)
            _lons = ma.masked_array(self.lons.data, latslonsmask)
            _latsCenter = ma.masked_array(self.latsCenter.data, centerMask)
            _lonsCenter = ma.masked_array(self.lonsCenter.data, centerMask)
            imgMask = np.repeat(centerMask[:,:,None], self.img.shape[2], 2)
            _img = ma.masked_array(self.img.data, imgMask)
            _elevation = ma.masked_array(self.elevation.data, centerMask)
            
            class attrs(object):
                lats=_lats
                lons=_lons
                latsCenter=_latsCenter
                lonsCenter=_lonsCenter
                img=_img
                elevation=_elevation
                        
            class MaskedMapping(object):                    
                @property
                def lats(self):
                    return self._sanitized(lambda s: s.lats)
                
                @property
                def lons(self):
                    return self._sanitized(lambda s: s.lons)
                
                @property
                def latsCenter(self):
                    return self._sanitized(lambda s: s.latsCenter)
                
                @property
                def lonsCenter(self):
                    return self._sanitized(lambda s: s.lonsCenter)
                
                @property
                def img(self):
                    return self._sanitized(lambda s: s.img)
                
                @property
                def elevation(self):
                    return self._sanitized(lambda s: s.elevation)
                
                def _sanitized(self, p):
                    self._sanitize(attrs)
                    return p(attrs)
                
            m = copy.copy(self)
            extend(m, MaskedMapping)
            m.isSanitized = False
            m.afterMasking = True
            return m
            
        def _sanitized(self, p):
            self._sanitize(super(SanitizedMapping, self))
            return p(super(SanitizedMapping, self))
            
        def _sanitize(self, attrs):
            if getattr(self, 'isSanitized', False):
                return
            afterMasking = getattr(self, 'afterMasking', False)
            
            _doSanitize(attrs.lats, attrs.lons, attrs.latsCenter, 
                        attrs.lonsCenter, attrs.img, attrs.elevation,
                        afterMasking=afterMasking)
            self.isSanitized = True
    
    SanitizedMapping.__doc__ = cls.__doc__
    
    return SanitizedMapping

@sanitize_data
@inherit_docs
class GenericMapping(ArrayImageMixin, BaseMapping):
    """
    A mapping consisting of precalculated latitudes/longitudes/elevation values.
    """
    
    def __init__(self, lats, lons, latsCenter, lonsCenter, elev, alti, img, cameraPosGCRS, photoTime, 
                 identifier, metadata=None):
        """
        
        :param ndarray lats: (h+1,w+1) in degrees
        :param ndarray lons: (h+1,w+1) in degrees
        :param ndarray latsCenter: (h,w) in degrees
        :param ndarray lonsCenter: (h,w) in degrees
        :param ndarray elev: (h,w) in degrees; can also be None         
        :param number alti: the altitude in km onto which the image was mapped (e.g. 110)
        :param ndarray img: uint8 or uint16 array of shape (h,w) for grayscale or (h,w,3) for RGB
        :param array-like cameraPosGCRS: [x,y,z] in km
        :param datetime.datetime photoTime:
        """
        h, w = img.shape[0], img.shape[1]
        assert lats.shape == lons.shape == (h+1, w+1)
        assert latsCenter.shape == lonsCenter.shape == (h, w)
        assert elev is None or elev.shape == (h, w)
        
        ArrayImageMixin.__init__(self, img)
        BaseMapping.__init__(self, alti, cameraPosGCRS, photoTime, identifier, metadata)
        self._lats = ma.masked_invalid(lats, copy=False) if not ma.isMA(lats) else lats
        self._lons = ma.masked_invalid(lons, copy=False) if not ma.isMA(lons) else lons
        self._latsCenter = ma.masked_invalid(latsCenter, copy=False) if not ma.isMA(latsCenter) else latsCenter
        self._lonsCenter = ma.masked_invalid(lonsCenter, copy=False) if not ma.isMA(lonsCenter) else lonsCenter
        self._elevation = ma.masked_invalid(elev, copy=False) if not ma.isMA(elev) else elev
        
    @property
    def lats(self):
        return self._lats
    
    @property
    def lons(self):
        return self._lons
    
    @property
    def latsCenter(self):
        return self._latsCenter
    
    @property
    def lonsCenter(self):
        return self._lonsCenter
    
    @property
    def elevation(self):
        return self._elevation
    
    def createResampled(self, lats, lons, latsCenter, lonsCenter, elevation, img):
        mapping = GenericMapping(lats, lons, latsCenter, lonsCenter, elevation, self.altitude, img, 
                                 self.cameraPosGCRS, self.photoTime, self.identifier,
                                 metadata=self.metadata)
        return mapping
    
    @staticmethod
    def fromMapping(mapping):
        """
        Create a :class:`GenericMapping` from the given mapping.
        
        Useful for removing additional data of a more specific mapping class.
        Note that this method can only be used if the image data is compatible
        to :class:`GenericMapping` (standard uint8 or uint16 grayscale/RGB arrays).        
        """
        m = GenericMapping(mapping.lats, mapping.lons, mapping.latsCenter, mapping.lonsCenter, 
                           mapping.elevation, mapping.altitude, mapping.img, 
                           mapping.cameraPosGCRS, mapping.photoTime, mapping.identifier, mapping.metadata)
        try:
            m.isSanitized = mapping.isSanitized
            m.afterMasking = mapping.afterMasking
        except:
            pass
        return m
# for sphinx: this avoids the class being displayed as an alias of SanitizedMapping
# The bases are displayed wrongly though.
GenericMapping.__name__ = 'GenericMapping'

class MappingCollection(object):
    def __init__(self, mappings, identifier, mayOverlap=True):
        """
        A collection of mappings for the same photo time (+- a few seconds).
        
        :type mappings: list of :class:`BaseMapping` objects
        :param string identifier: e.g. THEMIS-2012.01.01.12.30.59 
        :param mayOverlap: True if some mappings may overlap each other
                           This information is used for drawing polygons. When True,
                           then the pixels of all mappings are joined together, sorted
                           by elevation, and drawn in the same order (such that low-elevation
                           polygons of one mapping are overdrawn by higher-elevation polygons
                           of another overlapping mapping).
        """
        self._mappings = mappings
        self._identifier = identifier
        self._mayOverlap = mayOverlap
    
    @property
    def identifier(self):
        return self._identifier
    
    @property
    def empty(self):
        """
        Whether this collection contains no mappings.
        """
        return len(self.mappings) == 0
    
    @property
    def mappings(self):
        return self._mappings
    
    @property
    def mayOverlap(self):
        return self._mayOverlap
    
    def maskedByElevation(self, minElevation=10):
        mappings = [m.maskedByElevation(minElevation) for m in self.mappings]
        return MappingCollection(mappings, self.identifier, self.mayOverlap)
            
    @property
    def boundingBox(self):
        """
        The smallest bounding box containing the bounding boxes of all mappings.
        """
        boundingBoxes = [mapping.boundingBox for mapping in self.mappings]
        return BoundingBox.mergedBoundingBoxes(boundingBoxes)
    
    @property
    def photoTime(self):
        """
        The median of the photo times of all mappings.
        """
        times = sorted(m.photoTime for m in self.mappings)
        return times[len(times)//2]
    
    def __len__(self):
        return len(self._mappings)
               
@add_metaclass(ABCMeta)
class BaseMappingProvider(object):
    """
    Base class for all mapping providers.
    """
    
    def __init__(self, maxTimeOffset):
        """
        :param maxTimeOffset: in seconds
        """
        self.maxTimeOffset = maxTimeOffset
        
    @abstractproperty
    def range(self):
        """
        The dates of the first and last available mappings. 
        
        :rtype: datetime tuple (from, to)
        """
    
    @abstractmethod
    def contains(self, date):
        """
        Return True if there is a mapping for the given date 
        within +-maxTimeOffset.
        
        :param datetime date:
        :rtype: bool
        """
        
    def containsAny(self, dates):
        """
        Return True if there is a mapping for at least one of the given dates
        within +-maxTimeOffset.
        
        :param dates: list of datetime objects
        :rtype: bool
        """
        return any(self.contains(date) for date in dates)
        
    @abstractmethod
    def get(self, date):
        """
        Returns the mapping which is closest to the given date
        within +-maxTimeOffset.
        
        :param datetime date:
        :rtype: BaseMapping or MappingCollection
        :raise ValueError: when no mapping exists for the given date
        """
    
    @abstractmethod
    def getById(self, identifier):
        """
        Returns the mapping with the given identifier.
        
        :param string identifier:
        :rtype: BaseMapping or MappingCollection
        :raise ValueError: when no mapping with the given identifier exists
        """
    
    @abstractmethod
    def getSequence(self, dateBegin=None, dateEnd=None):
        """
        Returns a generator of mappings ordered by date for the given date range.
        If dateBegin and dateEnd are None, then all available mappings are returned.
        
        :param datetime dateBegin: 
        :param datetime dateEnd: inclusive
        :rtype: list of BaseMapping or MappingCollection objects
        """

def MaskByElevationProvider(provider, *args, **kw):
    """
    Wrap the given mapping provider by masking every returned mapping by elevation.    
    
    :param provider: the provider to wrap
    
    See :func:`BaseMapping.maskedByElevation` for masking parameters.
    """
    def mask(m):
        return m.maskedByElevation(*args, **kw)
    class MaskingProvider(object):          
        def get(self, *args, **kw):
            m = super(MaskingProvider, self).get(*args, **kw)
            return mask(m)
        
        def getById(self, *args, **kw):
            m = super(MaskingProvider, self).getById(*args, **kw)
            return mask(m)
            
        def getSequence(self, *args, **kw):
            m = super(MaskingProvider, self).getSequence(*args, **kw)
            return map(mask, m)
    
    provider = copy.copy(provider)  
    extend(provider, MaskingProvider)
    return provider

def inflatedEarthIntersection(cameraToPixelDirection, cameraPos, earthInflation=110, earthModel='wgs84'):
    """
    Return the intersection points with an inflated earth
    when shooting rays originating at `cameraPos` and 
    going in the direction `cameraToPixelDirection`.
    
    When `earthModel` is 'sphere' then the shape of the inflated earth
    is described by a sphere with radius `earthRadius + earthInflation`.
    
    When `earthModel` is 'wgs84' then the shape of the inflated earth
    is described by an ellipsoid with an equatorial axis of
    `wgs84A + earthInflation` and a polar axis of `wgs84B + earthInflation`
    where `wgs84A` and `wgs84B` refers to the 
    `World Geodetic System 84 <https://en.wikipedia.org/wiki/World_Geodetic_System>`_.
    
    :param cameraToPixelDirection: direction vectors in ICRS from camera location to pixel/sky location, shape (n,3)
    :param cameraPos: ndarray of xyz J2000 coordinates in km
    :param earthInflation: in km, how much to expand the earth when intersecting, use 0 for earth intersection
    :param earthModel: earth model, either 'wgs84' or 'sphere'
    """
    assert ((cameraToPixelDirection.ndim == 1 and cameraToPixelDirection.shape[0] == 3) or
            (cameraToPixelDirection.ndim == 2 and cameraToPixelDirection.shape[1] == 3))
    t0 = time.time()
    if earthModel == 'wgs84':
        a = wgs84A + earthInflation
        b = wgs84B + earthInflation
        intersectionAurora = ellipsoidLineIntersection(a, b, cameraPos, cameraToPixelDirection)
        
    elif earthModel == 'sphere':
        earthRadius = const.R_earth.to(u.km).value
        sphereRadius = earthRadius + earthInflation
        intersectionAurora = sphereLineIntersection(sphereRadius, cameraPos, cameraToPixelDirection)
        
    else:
        raise ValueError('unsupported earth model: ' + earthModel)
    print('inflatedEarthIntersection:', time.time()-t0, 's')    
    return intersectionAurora

class _SMMapping(GenericMapping):
    @property
    def cameraFootpoint(self):
        mlat, mlt = j2000ToMLatMLT([self.cameraPosGCRS], self.photoTime)
        smlon = mltToSmLon(mlt)
        return Location(mlat[0], smlon[0])

def convertMappingToSM(mapping):
    """
    Return a new mapping with the coordinates transformed
    to solar magnetic latitudes and longitudes.
    
    .. warning:: This function is not intended for regular use.
        If you need to access MLat/MLT coordinates, use
        the :attr:`BaseMapping.mLatMlt` and :attr:`BaseMapping.mLatMltCenter`
        attributes.
        
    See :func:`auromat.draw.drawStereographicMLatMLT` for an application
    of this function.
    
    :param BaseMapping mapping: 
        A mapping where latitudes and longitudes are given
        in the standard geodetic coordinate system.  
    :rtype: BaseMapping
    """
    mlat, mlt = mapping.mLatMlt
    mlat, mlt = mlat.data, mlt.data
    smlons = mltToSmLon(mlt)
    mlatCenter, mltCenter = mapping.mLatMltCenter
    mlatCenter, mltCenter = mlatCenter.data, mltCenter.data
    smlonsCenter = mltToSmLon(mltCenter)
    newMapping = _SMMapping(mlat, smlons, mlatCenter, smlonsCenter, 
                            mapping.elevation, mapping.altitude, 
                            mapping.img, mapping.cameraPosGCRS, 
                            mapping.photoTime, mapping.identifier)
    return newMapping

def convertSMMappingToGeo(mapping):
    """ Inverse operation to :func:`convertMappingToSM`. """
    smlats, smlons = mapping.lats.data, mapping.lons.data
    smlatsCenter, smlonsCenter = mapping.latsCenter.data, mapping.lonsCenter.data
    lats, lons = smToLatLon(smlats, smlons, mapping.photoTime)
    latsCenter, lonsCenter = smToLatLon(smlatsCenter, smlonsCenter, mapping.photoTime)    
    newMapping = GenericMapping(lats, lons, latsCenter, lonsCenter, 
                                mapping.elevation, mapping.altitude, 
                                mapping.img, mapping.cameraPosGCRS, 
                                mapping.photoTime, mapping.identifier)
    return newMapping



# Copyright European Space Agency, 2013

"""
This module provides various functions for visualizing mappings,
astrometric solutions, and distortion profiles.

The return value is nearly always a tuple consisting of
a matplotlib figure and axes object. This can then be further
modified or saved as an image file with :func:`saveFig`.
"""

from __future__ import division, absolute_import, print_function

from six import reraise, string_types
import os
import sys
import warnings
import datetime
from functools import partial
from auromat.util.coroutine import coroutine, throw

import numpy as np
import numpy.ma as ma
from astropy.coordinates.angles import Angle
import astropy.units as u
from scipy.spatial import Delaunay

import matplotlib as mpl
from matplotlib.ticker import FuncFormatter
from matplotlib.mlab import poly_between
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.collections import PolyCollection
from mpl_toolkits.basemap import Basemap
import brewer2mpl
from astropy.wcs.wcs import WCS
from auromat.fits import readHeader
import itertools
from matplotlib.patches import Polygon
from math import modf
blue_red = brewer2mpl.get_map('RdBu', 'Diverging', 11, reverse=True).mpl_colormap


from auromat.draw_helpers import overlapPolygons,\
    generatePolygonsFromMapping, \
    _generatePolygonsFromMappingOrCollection, _addFigureBottomTitle,\
    _convertMappingsToSM, _formatMLT, ensureContinuousPath,\
    loadFigImage, _saveFig, _getMplColors, _setMplColors, _circles, ColorMode
from auromat.resample import plateCarreeResolution, _resample
from auromat.utils import outline
import auromat.resources
import auromat.resample
import auromat.fits
from auromat.coordinates.geodesic import Location
from auromat.coordinates import geodesic, constellations
from auromat.mapping.mapping import BoundingBox
try:
    import auromat.util.lensdistortion
    import lensfunpy
except Exception as e:
    print(str(e))
    warnings.warn('lensfunpy not found, reduced functionality in auromat.draw')
    
DatePlotTitleFormat = '%Y-%m-%d %H:%M:%S.%f UTC'
DatePlotTitleWithoutUTCFormat = '%Y-%m-%d %H:%M:%S.%f'
            
def drawPlot(mapping):
    """
    Draws a single mapping onto a Longitude/Latitude plot equivalent
    to a rectilinear map projection.
    
    :rtype: tuple(Figure, Axes)
    """
    verts, colors = generatePolygonsFromMapping(mapping, ColorMode.matplotlib)
    
    if mapping.boundingBox.containsDiscontinuity:
        verts[:,:,1] = Angle((verts[:,:,1] + 180) * u.deg).wrap_at(180 * u.deg).degree
            
    fig, ax = plt.subplots()

    ax.set_xlabel('Longitude (deg)')
    ax.set_ylabel('Latitude (deg)')      

    verts = overlapPolygons(verts)
    
    verts = np.dstack((verts[:,:,1],verts[:,:,0])) # swap axes such that x=longitude

    coll = PolyCollection(verts, facecolors=colors, edgecolors='none', zorder=2,
                          rasterized=True)
    ax.add_collection(coll)
    ax.autoscale()
    
    if mapping.boundingBox.containsDiscontinuity:
        def fixAxis(x,pos):
            deg = Angle((x + 180) * u.deg).wrap_at(180 * u.deg).degree
            s = '{:g}'.format(deg) # removes trailing .0
            return s.replace('-', u'\u2212') # uses unicode minus (longer)
            # Note: this should rather make use of ScalarFormatter but I couldn't figure it out
        ax.xaxis.set_major_formatter(FuncFormatter(fixAxis))

    return fig, ax

def drawKmlImage(mapping):
    """
    Draws a mapping onto a figure without borders for use within Google Earth.
    The extents are identical to the bounding box of the mapping.
    
    Note: When saving, make sure to use `transparent=True` to make the areas
    outside the valid data transparent.
    
    :rtype: tuple(Figure, Axes)
    """
    boundingBox = mapping.boundingBox
    verts, colors = generatePolygonsFromMapping(mapping, ColorMode.matplotlib)
        
    if boundingBox.containsDiscontinuity:
        verts[:,:,1] = Angle((verts[:,:,1] + 180) * u.deg).wrap_at(180 * u.deg).degree
        lonWest = Angle((boundingBox.lonWest + 180) * u.deg).wrap_at(180 * u.deg).degree
        lonEast = Angle((boundingBox.lonEast + 180) * u.deg).wrap_at(180 * u.deg).degree
        boundingBox = BoundingBox(boundingBox.latSouth, lonWest, boundingBox.latNorth, lonEast)
        
    fig = plt.figure()
    fig.set_size_inches(1, 1)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    ax.set_autoscale_on(False)
    ax.set_xlim(boundingBox.lonWest, boundingBox.lonEast)
    ax.set_ylim(boundingBox.latSouth, boundingBox.latNorth)
    fig.add_axes(ax)

    verts = overlapPolygons(verts)
    
    verts = np.dstack((verts[:,:,1],verts[:,:,0])) # swap axes such that x=longitude

    coll = PolyCollection(verts, facecolors=colors, edgecolors='none') # expects [r,g,b]
    ax.add_collection(coll)
    
    return fig, ax

def drawStereographic(mappings, lat0=None, lon0=None, width=None, height=None, 
                      boundingBox=None,
                      sizeFactor=1, lineDelta=10,
                      labelsParallels=[True, False, False, False], labelsMeridians=[False, False, False, True],
                      fmtParallels=None, fmtMeridians=None, bottomTitle=None,
                      drawlsmask=True,
                      drawCities=False, cityAlpha=0.6, figWidth=8.0, figHeight=5.0, rasterized=True):
    """
    Draws one or more mappings onto a stereographic map.
    
    Note that when drawing multiple mappings, the resulting bounding box should
    span less than 180 degrees in longitude.
    
    :param mappings: either a single BaseMapping, a MappingCollection, or a list of
                     both types
    :param degree|None lat0: latitude of image center
    :param degree|None lon0: longitude of image center
    :param number|None width: width of map in km
    :param number|None height: height of map in km
    :param BoundingBox boundingBox: determine lat0,lon0,width,heigth from this bounding box
    :param number sizeFactor: how much to zoom out, 1 = everything just fits on the map
    :param int lineDelta: draw a parallel/meridian every lineDelta degrees
    :param labelsParallels: a boolean list [left,right,top,bottom] that determines on which sides
                            to draw labels for parallels
    :param labelsMeridians: a boolean list [left,right,top,bottom] that determines on which sides
                            to draw labels for meridians
    :param str|function fmtParallels: format string or function for parallel labels
    :param str|function fmtMeridians: format string or function for meridian labels
    :param str bottomTitle: adds a title below the figure
    :param bool drawlsmask: whether to draw land-sea mask
    :param bool drawCities: whether to draw city markers
    :param number cityAlpha: the alpha value with which to draw the city markers
    :param number figWidth: figure width in inches
    :param number figHeight: figure height in inches
    :param bool rasterized: whether to rasterize the polygon mesh when saved as SVG
    :rtype: tuple(Figure, Axes, Basemap)
    """
    if isinstance(mappings, list):
        vertsArr = []
        colorsArr = []
        for mapping in mappings:
            vertsArr_, colorsArr_ = _generatePolygonsFromMappingOrCollection(mapping)
            vertsArr.extend(vertsArr_)
            colorsArr.extend(colorsArr_)
        boundingBoxes = [mapping.boundingBox for mapping in mappings]
        photoTimes = [mapping.photoTime for mapping in mappings]
        photoTimeRange = (min(photoTimes), max(photoTimes))
    else:
        vertsArr, colorsArr = _generatePolygonsFromMappingOrCollection(mappings)
        boundingBoxes = [mappings.boundingBox]
        photoTimeRange = (mappings.photoTime, mappings.photoTime)
    
    if lat0 is None or lon0 is None or width is None or height is None:
        if boundingBox is None:
            boundingBox = BoundingBox.mergedBoundingBoxes(boundingBoxes)
        if lat0 is None:
            lat0 = boundingBox.center.lat
        if lon0 is None:
            lon0 = boundingBox.center.lon

        if width is None:
            width = boundingBox.size.width * 1.05 * sizeFactor
        if height is None:
            height = boundingBox.size.height * 1.05 * sizeFactor
    
    fig, ax, m = _prepareStereographicBasemap(figWidth, figHeight, lat0, lon0, width, height,
                                              lineDelta=lineDelta, labelsParallels=labelsParallels,
                                              labelsMeridians=labelsMeridians, drawlsmask=drawlsmask,
                                              fmtMeridians=fmtMeridians, fmtParallels=fmtParallels)

    if photoTimeRange[0] == photoTimeRange[1]:
        fig.suptitle(photoTimeRange[0].strftime(DatePlotTitleFormat))
    else:
        fig.suptitle(photoTimeRange[0].strftime(DatePlotTitleFormat) + ' - ' + 
                     photoTimeRange[1].strftime(DatePlotTitleFormat))
        
    if bottomTitle:
        _addFigureBottomTitle(fig, bottomTitle)
    
    _drawPolygons(m, ax, vertsArr, colorsArr, rasterized=rasterized)
    if drawCities:
        _drawCities(ax, m, alpha=cityAlpha)
    return fig, ax, m
    
def drawStereographicMLatMLT(mappings, *args, **kwargs):
    """
    Draws one or more mappings in the MLat/MLT coordinate system with stereographic projection.
    
    See :func:`drawStereographic` for parameter descriptions.
    
    :rtype: tuple(Figure, Axes, Basemap)
    """
    mappings = _convertMappingsToSM(mappings)
    
    bottomTitle = kwargs.pop('bottomTitle', 'MLat/MLT')
    fmtMeridians = kwargs.pop('fmtMeridians', partial(_formatMLT, format_='%H:%M'))
    # as we use a different reference frame we cannot draw any shapes based on the
    # standard geographic coordinates
    kwargs['drawCities'] = False
    kwargs['drawlsmask'] = False
    return drawStereographic(mappings, *args, fmtMeridians=fmtMeridians, bottomTitle=bottomTitle, **kwargs)

def drawMLatMLTPolar(mappings, lineDelta=10, boundingBox=None, figWidth=8.0, figHeight=5.0, rasterized=True):
    """
    Draws one or more mappings in the MLat/MLT coordinate system with 
    polar azimuthal equidistant projection.
    
    :param mappings: either a single BaseMapping, a MappingCollection, or a list of
                     both types
    :param int lineDelta: draw a parallel/meridian every lineDelta degrees
    :param BoundingBox boundingBox: determines the latitude range from this bounding box    
    :param number figWidth: figure width in inches
    :param number figHeight: figure height in inches
    :param bool rasterized: whether to rasterize the polygon mesh when saved as SVG
    :rtype: tuple(Figure, Axes, Basemap)
    """
  
    def coordsFn(mapping):
        mlat, mlt = mapping.mLatMlt
        # [0,24] -> [-180,180]
        smlon = mlt-12
        smlon /= (24/360)
        return mlat, smlon
  
    if isinstance(mappings, list):
        vertsArr = []
        colorsArr = []
        for mapping in mappings:
            vertsArr_, colorsArr_ = _generatePolygonsFromMappingOrCollection(mapping, coordsFn=coordsFn)
            vertsArr.extend(vertsArr_)
            colorsArr.extend(colorsArr_)
        photoTimes = [mapping.photoTime for mapping in mappings]
        photoTimeRange = (min(photoTimes), max(photoTimes))
    else:
        vertsArr, colorsArr = _generatePolygonsFromMappingOrCollection(mappings, coordsFn=coordsFn)
        photoTimeRange = (mappings.photoTime, mappings.photoTime)
    
    if boundingBox is None:
        latSouth = min(np.min(v[:,0,0]) for v in vertsArr)
        latNorth = max(np.max(v[:,0,0]) for v in vertsArr)
    else:
        latSouth = boundingBox.latSouth
        latNorth = boundingBox.latNorth
        
    fig, ax, m = _preparePolarBasemap(figWidth, figHeight, latSouth, latNorth, 
                                      lon_0=180, round_=True,
                                      drawlsmask=False, drawgrid=False)
    
    if photoTimeRange[0] == photoTimeRange[1]:
        fig.suptitle(photoTimeRange[0].strftime(DatePlotTitleFormat))
    else:
        fig.suptitle(photoTimeRange[0].strftime(DatePlotTitleFormat) + ' - ' +
                     photoTimeRange[1].strftime(DatePlotTitleFormat))

    # bottom title
    fig.text(0.5, 0.02, 'MLat/MLT', horizontalalignment='center')
    
    m.drawparallels(np.arange(-90,90,lineDelta))
    
    # draw latitude labels along 9h line
    if (latSouth+latNorth)/2 > 0: # north pole projection
        latonplot = lambda lat: m.boundinglat <= lat
    else: # south pole projection
        latonplot = lambda lat: lat <= m.boundinglat
    lon = -45
    for lat in filter(latonplot, range(-80, 81, 10)):
        x,y = m(lon,lat)
        x,y = ax.transLimits.transform((x,y))
        x += 0.03
        ax.text(x, y, str(lat), horizontalalignment='left', verticalalignment='center', 
                transform=ax.transAxes)
    
    m.drawmeridians(list(range(-180,180,45)), labels=[True, True, True, True], fmt=_formatMLT)
    
    fig._applyMeridiansParallelsColors = True
    
    _drawPolygons(m, ax, vertsArr, colorsArr, rasterized=rasterized)
    return fig, ax, m

def _prepareStereographicBasemap(figWidth, figHeight, lat0, lon0, mapWidth, mapHeight, lineDelta=10,
                                 drawlsmask=True, drawgrid=True,
                                 labelsParallels=[True,False,False,False], labelsMeridians=[False,False,False,True],
                                 fmtParallels=None, fmtMeridians=None):
    assert mapWidth > 0 and mapHeight > 0
    print('centering map at', lat0, lon0)
    print('sizing map at', mapWidth, mapHeight, 'km')

    fig = plt.figure(figsize=(figWidth,figHeight))
    ax = fig.add_axes([0.1,0.1,0.8,0.8])
    
    width = mapWidth*1000
    height = mapHeight*1000
    
    try:
        m = Basemap(width=width, height=height, # in meters
                    resolution='l', projection='stere', ellps='WGS84',
                    lat_0=lat0, lon_0=lon0, ax=ax)
    except:
        # add some debug information for reproducing the issue
        _, e, tb = sys.exc_info()
        new_exc = Exception('{}: {}; width={}, height={}, lat_0={}, lon_0={}'.\
                            format(e.__class__.__name__, e, width, height, lat0, lon0))
        reraise(new_exc.__class__, new_exc, tb)
        
    
    if drawlsmask:
        m.drawlsmask(ocean_color='0.8', land_color='0.6', lakes=False, resolution='l',grid=1.25) # lower res: resolution='c'
    
    if drawgrid:
        # TODO find good line frequency automatically
        kw = {}
        if fmtParallels:
            kw['fmt'] = fmtParallels  
        m.drawparallels(np.arange(-80.,81.,lineDelta), labels=labelsParallels, **kw) # labelstyle="+/-"
        
        kw = {}
        if fmtMeridians:
            kw['fmt'] = fmtMeridians
        m.drawmeridians(np.arange(-180.,181.,lineDelta), labels=labelsMeridians, **kw)
        
        fig._applyMeridiansParallelsColors = True
            
    return fig, ax, m

def _preparePolarBasemap(figWidth, figHeight, latSouth, latNorth, lineDelta=10, lon_0=180,
                         drawlsmask=True, drawgrid=True, round_=False,
                         labelsParallels=[True,False,False,False], labelsMeridians=[False,False,False,True]):
    """
    
    :param lon_0: lon_0 is at 6-o'clock
    """

    fig = plt.figure(figsize=(figWidth,figHeight))
    ax = fig.add_axes([0.1,0.1,0.8,0.8])
        
    if (latSouth+latNorth)/2 > 0:
        # north pole
        projection='npaeqd'
        boundinglat = latSouth - 5
    else:
        # south pole
        projection='spaeqd'
        boundinglat = latNorth + 5
        
    m = Basemap(projection=projection, boundinglat=boundinglat, lon_0=lon_0,
                round=round_, resolution='l', ax=ax)

    if round_:
        # note that the actual color is set in draw_helpers.setMplColors
        m.drawmapboundary(fill_color='none')
        # see draw_utils._setMplColors
        fig._applyRoundPolarMapColors = True
        
    if drawlsmask:
        m.drawlsmask(ocean_color='0.8', land_color='0.6', lakes=False, resolution='l',grid=1.25) # lower res: resolution='c' 
    
    if drawgrid:
        # TODO find good line frequency automatically
        m.drawparallels(np.arange(-80.,81.,lineDelta),labels=labelsParallels) # labelstyle="+/-"
        m.drawmeridians(np.arange(-180.,181.,lineDelta),labels=labelsMeridians)
        fig._applyMeridiansParallelsColors = True
    
    return fig, ax, m

def _drawCities(ax, m, shapefile=None, color='red', alpha=0.6, moreCities=True):
    scale = '10' if moreCities else '50'
    if shapefile is None:
        resPath = os.path.dirname(auromat.resources.__file__)
        shapefile = os.path.join(resPath, 'ne_' + scale + 'm_populated_places_simple')
    m.readshapefile(shapefile, 'cities')
    natscale = np.array([i['natscale'] for i in m.cities_info])
    x,y = list(zip(*m.cities))
    xy = np.transpose([x,y])
    c = np.logical_and(m.xmin < xy[:,0], 
                       np.logical_and(xy[:,0] < m.xmax, 
                                      np.logical_and(m.ymin < xy[:,1], xy[:,1] < m.ymax)))
    xy = xy[c]
    natscale = natscale[c]/10
    
    s = ax.scatter(xy[:,0], xy[:,1], natscale, color, marker='o', edgecolors='none', zorder=10, alpha=alpha)
    # so that it can be accessed when outputting as svg
    s.set_gid('cities')
    
def _drawPolygons(m, ax, verts, colors, rasterized=True):
    """
    Draws one or more polygon meshes on a map.
    
    :param verts: list of (n, 4, 2) arrays in degrees (:,:,0 = lat)
    :param colors: list of (n, 3) arrays in [0,1]
    """    
    # see http://matplotlib.org/basemap/users/mapcoords.html
    verts_ = np.concatenate(verts)
    colors_ = np.concatenate(colors)
    if len(verts_) == 0:
        return
    
    lats = verts_[:,:,0].ravel()
    lons = verts_[:,:,1].ravel()
    xpt, ypt = m(lons,lats)
    verts_ = np.dstack((xpt.reshape(-1,4),ypt.reshape(-1,4)))

    verts_ = overlapPolygons(verts_)                                            
    coll = PolyCollection(verts_, facecolors=colors_, edgecolors='none', zorder=2,
                          rasterized=rasterized) # expects [r,g,b]
    ax.add_collection(coll)

def drawHorizon(astrometryMapping, color='blue', channel=None, lineThickness=2,
                useMappingAltitude=False, figax=None):
    """
    Draws the earth horizon of an unresampled astrometry-based mapping
    on top of the underlying image of the mapping.
    
    :param astrometryMapping: the unresampled mapping
    :type astrometryMapping: auromat.mapping.astrometry.BaseAstrometryMapping
    :param color: earth horizon line color (matplotlib color spec)
    :param str channel: if given (r|g|b) then draw only the given channel of the image
    :param lineThickness: line width
    :param figax: a (Figure,Axes) tuple which shall be used instead of creating a new one;
                  if given, then the mapping image is not drawn
    :rtype: tuple(Figure,Axes)
    """
    if figax:
        fig, ax = figax
    else:
        im = astrometryMapping.rgb_unmasked
        if channel:
            if channel == 'red':
                im = im[:,:,0]
            elif channel == 'green':
                im = im[:,:,1]
            elif channel == 'blue':
                im = im[:,:,2]
        fig, ax = loadFigImage(im)
    
    if astrometryMapping.isConsistent():
        if useMappingAltitude:
            intersection = astrometryMapping.intersectionInflatedCenter
        else:
            intersection = astrometryMapping.intersectsEarth
        
        curve = _getIntersectionCurve(intersection)
        ax.plot(curve[:,0], curve[:,1], color=color, lw=lineThickness)   

    return fig, ax

def _getIntersectionCurve(intersection):
    """
    :param intersection: either (h+1,w+1,3) with intersection points or as boolean intersects array (h+1,w+1)
                         Note that the array must include both intersected and non-intersected elements. 
    :rtype: (n,2) ndarray of curve points
    """
    assert (intersection.dtype == bool and intersection.ndim == 2) or (intersection.ndim == 3)
 
    if intersection.dtype == bool:
        intersects = intersection
    else:
        intersects = ~np.isnan(intersection[:,:,0])
        
    outl = outline(intersects)
    # filter out points along the image border
    yfilt = np.logical_and(0 < outl[:,1], outl[:,1] < intersection.shape[0]-1)
    xfilt = np.logical_and(0 < outl[:,0], outl[:,0] < intersection.shape[1]-1)
    curve = outl[np.logical_and(yfilt, xfilt)]
                
    # Now the curve might end up as two logical path segments, because
    # the initial start of the path may have been in the middle of the curve,
    # depending on where findContours of OpenCV starts.
    # To correct that, we find the jump and swap both curve parts.
    curve = ensureContinuousPath(curve)
    return curve

def drawHistogram(hist, vlines=[], xlabel=None, ylabel=None, **kw):
    """
    Draws a histogram in line-style with unspecified bin boundaries.
    
    :param hist: a one-dimensional array of histogram y-values 
    :param vlines: a list of (value,color) tuples describing vertical lines to draw
    :param str xlabel: label of x axis
    :param str ylabel: label of y axis
    :rtype: tuple(Figure,Axes)
    """
    bins = np.arange(len(hist))

    fig, ax = drawLinePlot(bins, hist, xlabel, ylabel, **kw)
    ax.set_xlim([0, bins[-1]])
    
    for (i,color) in vlines:
        ax.axvline(i, color=color)
        
    return fig, ax

def drawHeatmaps(mapping, pathPrefix='heat_', xlabel='Image Width (px)', ylabel='Image Height (px)',
                 bgcolor='white', widthPx=None, dpi=None):
    """
    Draws latitude, longitude, elevation, azimuth (if available, e.g. MIRACLE mappings)
    as blue-red heatmaps for the given mapping.
    
    The resulting filenames are:
    
    - `<pathPrefix>lats.png`
    - `<pathPrefix>lons.png`
    - `<pathPrefix>elevation.png`
    - `<pathPrefix>azimuth.png`
    
    :param str pathPrefix: path prefix for the images that are created
    :param bgcolor: plot background color (matplotlib color spec)
    :param widthPx: width in pixels of the saved images
    :param dpi: dpi of the saved images
    """
    save = partial(saveFig, bgcolor=bgcolor, widthPx=widthPx, dpi=dpi)
    save(pathPrefix + 'lats.png', drawHeatmap(mapping.lats, cbLabel='Latitude (deg)', 
                                              xlabel=xlabel, ylabel=ylabel))
    save(pathPrefix + 'lons.png', drawHeatmap(mapping.lons, cbLabel='Longitude (deg)',
                                              xlabel=xlabel, ylabel=ylabel))
    save(pathPrefix + 'elevation.png', drawHeatmap(mapping.elevation, cbLabel='Elevation angle (deg)', 
                                                   xlabel=xlabel, ylabel=ylabel))
    
    try: # not all mappings have azimuth data
        save(pathPrefix + 'azimuth.png', drawHeatmap(mapping.azimuthCenter, xlabel=xlabel, ylabel=ylabel))
    except:
        pass

def drawHeatmap(data, cbLabel=None, xlabel=None, ylabel=None, figax=None, alpha=1.0):
    """
    Draws a 2d-data array as blue-red heatmap.
    
    :param data: 2D array, can contains NaN's and/or be a masked numpy array
    :param cbLabel: data label, written alongside the colorbar
    :param figax: a (Figure,Axes) tuple to reuse
    :param alpha: alpha value of the drawn heatmap
    :rtype: tuple(Figure,Axes)
    """
    data = ma.masked_invalid(data, copy=False)
    fig, ax = plt.subplots() if figax is None else figax
    im = ax.imshow(data, cmap=blue_red, alpha=alpha)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cb = fig.colorbar(im, cax=cax)
    if cbLabel is not None:
        cb.set_label(cbLabel)
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    # TODO fix colors
#    _setMplColors(None, cax, colors)
    return fig, ax

@coroutine
def drawScanLinesCo(outputFile, widthPx=None, dpi=None, transparent=False, bgcolor='black',
                    lat0=None, lon0=None, mapWidth=None, mapHeight=None,
                    fmtParallels=None, fmtMeridians=None, drawlsmask=True, bottomTitle=None,
                    labelsParallels=[True, False, False, False], labelsMeridians=[False, False, False, True],
                    lineDelta=10, figWidth=8.0, figHeight=5.0,
                    arcsecPerPx=100, lineWidthFactor=1):
    """
    Draws a scanline from each given mapping such that the resulting stereographic map
    gives a temporal and spatial varying overview of all mappings.
    Scanlines are spherical rectangles perpendicular to the 'flying' direction of the mapping
    centroids when seen on a map.
    
    Note that the centroid of each mapping becomes the center of each corresponding scanline, 
    therefore the sequence should be masked by elevation before-hand with a sensible value.
    
    The input to this coroutine can either be a sequence of unresampled mappings or
    a dictionary containing `mapping.properties` in the 'props' key and the masked and
    resampled mapping in the 'mapping' key. In the latter case, the `arcsecPerPx` parameter
    is ignored.
    
    Example:
    
    >>> from auromat.util.coroutine import broadcast
    >>> seq = ...retrieve mapping sequence
    >>> broadcast(seq, drawScanLinesCo('test.png'))
    
    :param number arcsecPerPx: spherical resolution used for resampling
    :param number lineWidthFactor: how much to widen the scanline width relative to the
                                   the computed width derived from the first and second mapping;
                                   useful if mappings in-between the sequence are missing
    
    For all other parameters, see :func:`saveFig` and :func:`drawStereographic`.
    """
    try:
        currentMapping = yield
    except GeneratorExit:
        raise ValueError('mapping sequence too short, need at least 2 mappings')
    
    try:
        currentMappingProps = currentMapping['props']
        currentMapping = currentMapping['mapping']
    except (KeyError,TypeError):
        currentMappingProps = currentMapping.properties
        pxPerDeg = plateCarreeResolution(currentMappingProps.boundingBox, arcsecPerPx)
        currentMapping = auromat.resample.resample(currentMapping, pxPerDeg=pxPerDeg)
    
    bb = currentMappingProps.boundingBox
    # constant height of rotated scanline box
    # it is bigger than needed, just to be sure that it contains as much as possible
    height = geodesic.distance(bb.topLeft, bb.bottomRight)*1.5
    
    # constant width of rotated bounding box
    # determined from first and second mapping
    width = None
    
         
    # the amount of azimuthCenter to add to the camera footpoint azimuthCenter to face the mapping centroid   
    deltaAzimuth = None
    # the distance between the camera footpoint and the mapping centroid
    deltaDistance = None
    
    # The reason for calculating the centroid azimuthCenter track from the camera footpoints
    # is smoothness. The direct calculation of azimuths from two centroids is too noisy
    # and cannot be used for creating the scan lines.
    # The assumption made is that the camera doesn't move (tilt etc.) throughout the sequence.
    
    photoTimes = []
    lineBoundingBoxes = []
    centroids = []
    azimuths = []
    maxHeight = 0 # maximum measured height across all scanlines 
    
    vertsArr = []
    colorsArr = []
    
    done = False
    while not done:
        try:
            nextMapping = yield
        except GeneratorExit:
            done = True
            if not photoTimes:
                raise ValueError('mapping sequence too short, need at least 2 mappings')
        
        photoTimes.append(currentMappingProps.photoTime)
        centroids.append(currentMappingProps.centroid)
        
        if not done:
            try:
                nextMappingProps = nextMapping['props']
                nextMapping = nextMapping['mapping']
            except (KeyError,TypeError):
                nextMappingProps = nextMapping.properties
                # Important: Resampling is done after all values like centroid etc.
                # have been stored using the non-resampled mapping (higher accuracy).
                # For drawing, we need to resample though to reduce polygon counts.
                
                # Note that we use the same pxPerDeg for all mappings to stay on the same grid
                nextMapping = auromat.resample.resample(nextMapping, pxPerDeg=pxPerDeg)
                
            azCamFoot = geodesic.course(currentMappingProps.cameraFootpoint, nextMappingProps.cameraFootpoint)
        
        if width is None:
            # The width must be at least as big such that a polygon/pixel
            # can fit inside, no matter how the scanline is oriented in regards
            # to the regular lat/lon grid. The diagonal size of the first polygon
            # multiplied by 3 is used to approximate this. The factor was determined
            # by trial-and-error until no polygons of any scanline were left out
            # anymore. The exact value would have to use the maximum polygon size across
            # the scanlines of all mappings, which would require 2 passes and hence
            # double the time. As this plot merely serves for overview purposes, 
            # the approximation is considered good enough.
            verts, _ = generatePolygonsFromMapping(currentMapping)
            p1 = Location(verts[0][0][0], verts[0][0][1])
            p2 = Location(verts[0][2][0], verts[0][2][1])
            diagonalDistancePixel = geodesic.distance(p1, p2)
            
            consecutiveMappingDistance = geodesic.distance(currentMappingProps.centroid, nextMappingProps.centroid)
            width = max(diagonalDistancePixel*3, consecutiveMappingDistance) * lineWidthFactor
            
            deltaDistance = geodesic.distance(currentMappingProps.cameraFootpoint, currentMappingProps.centroid)
            
            azCamFootToCentroid = geodesic.course(currentMappingProps.cameraFootpoint, currentMappingProps.centroid)
            deltaAzimuth = azCamFoot - azCamFootToCentroid
            
        if not done:
            azCamFootToCentroid = azCamFoot - deltaAzimuth
            centroidFromAzCurrent = geodesic.destination(currentMappingProps.cameraFootpoint, azCamFootToCentroid, deltaDistance)
            centroidFromAzNext = geodesic.destination(nextMappingProps.cameraFootpoint, azCamFootToCentroid, deltaDistance)
            az = geodesic.course(centroidFromAzCurrent, centroidFromAzNext)
        else:
            # reuse the azimuth from the previous two frames
            # it shouldn't change that much from the last one
            pass
        
        azimuths.append(az)
                    
        middleRight = geodesic.destination(currentMappingProps.centroid, az, width/2)
        middleLeft = geodesic.destination(currentMappingProps.centroid, az+180, width/2)
        
        topLeft = geodesic.destination(middleLeft, az-90, height/2)
        bottomLeft = geodesic.destination(middleLeft, az+90, height/2)
        topRight = geodesic.destination(middleRight, az-90, height/2)
        bottomRight = geodesic.destination(middleRight, az+90, height/2)
        
        # skip last point of each geodesic line so that there are no duplicate points
        polygon = np.concatenate((geodesic.line(topLeft, topRight)[:-1],
                                  geodesic.line(topRight, bottomRight)[:-1],
                                  geodesic.line(bottomRight, bottomLeft)[:-1],
                                  geodesic.line(bottomLeft, topLeft)[:-1]))
        
        currentMapping = currentMapping.maskedByPolygon(polygon)
        
        lineBoundingBoxes.append(currentMapping.boundingBox)
        
        # approximate real height of scanline
        # the commented code uses OpenCV
#        rect = cv.minAreaRect(np.require(currentMapping.outline, np.float32, 'C').reshape(-1,1,2))
#        box = cv.cv.BoxPoints(rect)
#        distance1 = geodesic.distance(Location(*box[0]), Location(*box[1]))
#        distance2 = geodesic.distance(Location(*box[1]), Location(*box[2]))
#        currentHeight = max(distance1, distance2)
        # Using minAreaRect would be the most accurate, but currently only OpenCV
        # includes that as a function. We approximate it by using the diagonal
        # of the bounding box. This may lead to slightly larger heights than the actual heights.
        currentHeight = geodesic.distance(currentMapping.boundingBox.topLeft, 
                                          currentMapping.boundingBox.bottomRight)
        if currentHeight > maxHeight:
            maxHeight = currentHeight

        verts, colors = generatePolygonsFromMapping(currentMapping, ColorMode.matplotlib)
        vertsArr.append(verts)
        colorsArr.append(colors)
        
#        print('vertsarr:', sum(a.nbytes/1024/1024 for a in vertsArr), 'MB')
#        print('colorsarr:', sum(a.nbytes/1024/1024 for a in colorsArr), 'MB')
        
        currentMapping = nextMapping
        currentMappingProps = nextMappingProps
        
    debug = False
    
    if len(vertsArr) == 0:
        raise ValueError('mapping sequence too short')
                    
    # calculate time axis coordinates
    # the ticks are distributed with the assumption that there are no big gaps
    # in regards to available mappings over the sequence time range
    
    # first, the axis line itself
    mappingCount = len(vertsArr)
    tickCount = 4
    axisLineResolution = 10
    axisLinePointCount = mappingCount//axisLineResolution
    axisLinePointCount = max(tickCount, axisLinePointCount)
    axisLineDistance = maxHeight/2*1.1 
    lineIndices = np.round(np.linspace(0, mappingCount-1, axisLinePointCount)).astype(np.int)

    axisLine = [geodesic.destination(centroids[i], azimuths[i]-90, axisLineDistance) 
                for i in lineIndices]
    
    # then, ticks and labels
    labelDistance = axisLineDistance*1.2
    tickIndices = np.round(np.linspace(0, mappingCount-1, tickCount)).astype(np.int)
    tickLines = [(geodesic.destination(centroids[i], azimuths[i]-90, axisLineDistance),
                  geodesic.destination(centroids[i], azimuths[i]-90, axisLineDistance*1.04))
                 for i in tickIndices]
    tickLabels = [(geodesic.destination(centroids[i], azimuths[i]-90, labelDistance),
                   photoTimes[i])
                  for i in tickIndices]

    estimatedLabelEdge = labelDistance*1.1
    tickLabelsEdges = [geodesic.destination(centroids[i], azimuths[i]-90, estimatedLabelEdge)           
                       for i in tickIndices]
    timeAxisBoundingBox = BoundingBox.minimumBoundingBox([[p.lat,p.lon] for p in tickLabelsEdges])
    
    # TODO account for text width of first/last label 
    
    boundingBoxes = lineBoundingBoxes + [timeAxisBoundingBox]
    
    # TODO duplicate code (see drawStereographic)
#        if lat0 is None or lon0 is None or mapWidth is None or mapHeight is None:
    
    combinedBoundingBox = BoundingBox.mergedBoundingBoxes(boundingBoxes)
    if lat0 is None or debug:
        lat0 = combinedBoundingBox.center.lat
    if lon0 is None or debug:
        lon0 = combinedBoundingBox.center.lon

    if mapWidth is None or debug:
        mapWidth = combinedBoundingBox.size.width * 1.05
    if mapHeight is None or debug:
        mapHeight = combinedBoundingBox.size.height * 1.05
    
    # finally, draw everything
    fig, ax, m = _prepareStereographicBasemap(figWidth, figHeight, lat0, lon0, mapWidth, mapHeight,
                                 lineDelta=lineDelta, labelsParallels=labelsParallels, labelsMeridians=labelsMeridians,
                                 drawlsmask=drawlsmask, fmtParallels=fmtParallels, fmtMeridians=fmtMeridians)
    
    photoTimeRange = (min(photoTimes), max(photoTimes))

    if photoTimeRange[0] == photoTimeRange[1]:
        fig.suptitle(photoTimeRange[0].strftime(DatePlotTitleFormat))
    else:
        fig.suptitle(photoTimeRange[0].strftime(DatePlotTitleWithoutUTCFormat) + ' - ' + 
                     photoTimeRange[1].strftime(DatePlotTitleFormat))
        
    if bottomTitle:
        _addFigureBottomTitle(fig, bottomTitle)
    
    _drawPolygons(m, ax, vertsArr, colorsArr)
    
    if mappingCount > 0:
        x, y = m([p.lon for p in axisLine], [p.lat for p in axisLine])
        ax.plot(x, y)

        for (p1,p2) in tickLines:
            lats = [p1.lat, p2.lat]
            lons = [p1.lon, p2.lon]
            x, y = m(lons, lats)
            ax.plot(x, y)
        
        for (bottomCenter, date) in tickLabels:
            x, y = m(bottomCenter.lon, bottomCenter.lat)
            timeStr = date.strftime('%H:%M:%S')
            ax.text(x, y, timeStr, ha='center', va='bottom') #fontsize=12
    
    saveFig(outputFile, (fig,ax), bgcolor=bgcolor, transparent=transparent, widthPx=widthPx, dpi=dpi)

@coroutine
def drawScanLinesMLatMLTCo(*args, **kwargs):
    """
    As :func:`drawScanLinesCo` but with MLat/MLT as coordinate system.
    """
    fmtMeridians = kwargs.pop('fmtMeridians', partial(_formatMLT, format_='%H:%M'))
    bottomTitle = kwargs.pop('bottomTitle', 'MLat/MLT')
    # as we use a different reference frame we cannot draw any shapes based on the
    # standard geographic coordinates
    kwargs['drawlsmask'] = False
    drawFn = drawScanLinesCo(*args, fmtMeridians=fmtMeridians, bottomTitle=bottomTitle, **kwargs)
    
    try:
        while True:
            try:
                drawFn.send(_convertMappingToSM((yield)))
            except GeneratorExit:
                break
        
        drawFn.close()
    except:
        throw(drawFn, *sys.exc_info()) # will raise back to us    

def _convertMappingToSM(mapping):
    try:
        return {'props': mapping['props_sm'],
                'mapping': mapping['mapping_sm']}
    except KeyError:
        return _convertMappingsToSM(mapping)
        
@coroutine
def drawAzimuthPlotsCo(outputFileAzCentroid=None, outputFileAzCentroidFromCam=None,
                       outputFileAzCamFootToCentroid=None,
                       outputFileLatLonCentroid=None, outputFileLatLonCentroidFromAz=None,
                       outputFileLatLonCamFoot=None,
                       widthPx=None, dpi=None, transparent=False, bgcolor='black'):
    """
    Draws various plots based on the centroid of each mapping.
    
    :param outputFileAzCentroid: azimuth ('flying' direction) of the centroids 
    :param outputFileAzCentroidFromCam: azimuth ('flying' direction) of the recalculated centroids
                                        (using camera footpoints) 
    :param outputFileLatLonCentroid: latitude-longitude plot of centroids
    :param outputFileLatLonCentroidFromAz: latitude-longitude plot of recalculated centroids
                                           (using camera footpoints) 
    :param outputFileLatLonCamFoot: latitude-longitude plot of camera footpoints
    
    See :func:`saveFig` for other parameters.
    """
    try:
        mapping = yield
    except GeneratorExit:
        raise ValueError('mapping sequence too short')
    
    try:
        currentMappingProps = mapping['props']
    except (KeyError,TypeError):
        currentMappingProps = mapping.properties
    del mapping
            
    # the amount of azimuthCenter to add to the camera footpoint azimuthCenter to face the mapping centroid   
    deltaAzimuth = None
    # the distance between the camera footpoint and the mapping centroid
    deltaDistance = None
    
    # The reason for calculating the centroid azimuthCenter track from the camera footpoints
    # is smoothness. The direct calculation of azimuths from two centroids is too noisy
    # and cannot be used for creating the scan lines.
    
    boundingBoxes = []
    camFootpoints = []
    centroids = []
    centroidsFromAz = []
    azimuthsCamFootToCentroid = []
    azimuthsCentroid = []
    azimuthsCentroidFromCam = []
    
    photoTimes = []
        
    while True:
        try:
            mapping = yield
        except GeneratorExit:
            break
        
        try:
            nextMappingProps = mapping['props']
        except (KeyError,TypeError):
            nextMappingProps = mapping.properties
        del mapping
        
        boundingBoxes.append(currentMappingProps.boundingBox)
        centroids.append(currentMappingProps.centroid)
        photoTimes.append(currentMappingProps.photoTime)
        camFootpoints.append(currentMappingProps.cameraFootpoint)
        
        azCamFoot = geodesic.course(currentMappingProps.cameraFootpoint, nextMappingProps.cameraFootpoint)
        
        if deltaDistance is None:
            deltaDistance = geodesic.distance(currentMappingProps.cameraFootpoint, currentMappingProps.centroid)
            
            azCamFootToCentroid = geodesic.course(currentMappingProps.cameraFootpoint, currentMappingProps.centroid)
            deltaAzimuth = azCamFoot - azCamFootToCentroid
        
        
        # B. determine azimuthCenter using centroid
        azimuthsCentroid.append(geodesic.course(currentMappingProps.centroid, nextMappingProps.centroid))
        
        # C. determine azimuthCenter using recalculated centroid (from camera footpoint)
        azCamFootToCentroid = azCamFoot - deltaAzimuth # TODO check, possibly wrap
        centroidFromAzCurrent = geodesic.destination(currentMappingProps.cameraFootpoint, azCamFootToCentroid, deltaDistance)
        centroidFromAzNext = geodesic.destination(nextMappingProps.cameraFootpoint, azCamFootToCentroid, deltaDistance)
        centroidsFromAz.append(centroidFromAzCurrent)
        azimuthsCentroidFromCam.append(geodesic.course(centroidFromAzCurrent, centroidFromAzNext))
        azimuthsCamFootToCentroid.append(azCamFootToCentroid)
        
        currentMappingProps = nextMappingProps
            
    save = partial(saveFig, bgcolor=bgcolor, transparent=transparent, widthPx=widthPx, dpi=dpi)

    # B. draw time-azimuthCenter plot using centroids
    if outputFileAzCentroid:
        figax = drawLinePlot(photoTimes, azimuthsCentroid, 
                              xlabel='Time', 
                              ylabel='Azimuth ($^\circ$) using centroid')
        save(outputFileAzCentroid, figax)
    
    # C. draw time-azimuthCenter plot using recalculated centroids
    if outputFileAzCentroidFromCam:
        figax = drawLinePlot(photoTimes, azimuthsCentroidFromCam, 
                              xlabel='Time', 
                              ylabel='Azimuth ($^\circ$) using recalculated centroid')
        save(outputFileAzCentroidFromCam, figax)
                
    # B2. draw lat-lon line plot using centroids
    if outputFileLatLonCentroid:
        lats = [c.lat for c in centroids]
        lons = [c.lon for c in centroids]
        figax = drawLinePlot(lons, lats, 
                              xlabel='Longitude ($^\circ$) of centroid', 
                              ylabel='Latitude ($^\circ$) of centroid')
        save(outputFileLatLonCentroid, figax)
    
    # C2. draw lat-lon line plot using recalculated centroids
    if outputFileLatLonCentroidFromAz:
        lats = [c.lat for c in centroidsFromAz]
        lons = [c.lon for c in centroidsFromAz]
        figax = drawLinePlot(lons, lats, 
                              xlabel='Longitude ($^\circ$) of recalculated centroid', 
                              ylabel='Latitude ($^\circ$) of recalculated centroid')
        save(outputFileLatLonCentroidFromAz, figax)
    
    # draw lat-lon line plot of camera footpoints
    if outputFileLatLonCamFoot:
        lats = [c.lat for c in camFootpoints]
        lons = [c.lon for c in camFootpoints]
        figax = drawLinePlot(lons, lats, 
                              xlabel='Longitude ($^\circ$) of camera footpoint', 
                              ylabel='Latitude ($^\circ$) of camera footpoint')
        save(outputFileLatLonCamFoot, figax)

def drawLinePlot(x, y, xlabel, ylabel, title=None, linecolor=None, linewidth=None, retline=False, **kw):
    """
    Draws a line plot.
    
    :param array-like x: x values, can be datetime 
    :param array-like y: y values
    :param str xlabel: label of x axis
    :param str ylabel: label of y axis
    :param str title: plot title
    :param linecolor: line color (matplotlib color spec)
    :param linewidth: line width
    :param bool retline: if True, then the Line object is returned as last tuple element
    :param kw: additional keywords are handed over to matplotlib's `ax.plot` function
    :rtype: tuple (Figure,Axes[,Line])
    """
    fig,ax = plt.subplots()
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    if title:
        ax.set_title(title)
    if linecolor is not None:
        kw['color'] = linecolor
        
    if len(x)>0 and isinstance(x[0], datetime.datetime): 
        x_ = mpl.dates.date2num(x)
        line, = ax.plot_date(x_, y, fmt='b-', **kw)
        if max(d.microsecond for d in x) == 0:
            ax.xaxis.get_major_formatter().scaled[1/(24.*60.)] = '%H:%M:%S' # don't show float seconds
        fig.autofmt_xdate()
    else:    
        line, = ax.plot(x, y, **kw)
        
    if linewidth:
        line.set_linewidth(linewidth)
        
    if retline:
        return fig, ax, line
    else:
        return fig, ax
    
def _drawScatterPlot(x, y, xlabel, ylabel, title=None, pointcolor=None, retscatter=False, **kw):
    fig,ax = plt.subplots()
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    if title:
        ax.set_title(title)
    scatter = ax.scatter(x, y, color=pointcolor, **kw)
    if retscatter:
        return fig, ax, scatter
    else:
        return fig, ax

def drawLensDistortionDisplacementPlot(mod):
    """
    Draws the displacement in pixels caused by lens distortion as a 2D heatmap.
    
    :param lensfunpy.Modifier mod: the Modifier object
    :rtype: tuple(Figure,Axes) 
    """
    distances = auromat.util.lensdistortion.lensDistortionPixelDistances(mod=mod)
    
    calib = mod.lens.interpolate_distortion(mod.focal_length)
    model, params = _formatDistortionParams(calib)
    
    cbLabel = 'Displacement (px)'
    fig, ax = drawHeatmap(distances, cbLabel=cbLabel)
    
    ax.set_xlabel('Image Width (px)')
    ax.set_ylabel('Image Height (px)')
    ax.set_title(model + '(' + params + ')')
    
    return fig, ax
    
def drawLensDistortionDerivativePlot(mod):
    """
    Draws the lens distortion derivative.
    This visualizes barrel and pincushion distortion over the sensor radius.
    
    :param lensfunpy.Modifier mod: the Modifier object
    :rtype: tuple(Figure,Axes)
    """
    # derivatives of (rd - ru)/ru
    # see http://sourceforge.net/p/lensfun/mailman/message/32205273/
    def ptlens(ru, a, b, c):
        return 3*a*ru**2 + 2*b*ru + c
    
    def poly3(ru, k1):
        return 2*k1*ru
    
    def poly5(ru, k1, k2):
        return 2*k1*ru + 4*k2*ru**3
    
    # get the internal models interpolated for the given focal length
    calib = mod.lens.interpolate_distortion(mod.focal_length)
    
    if calib.model == lensfunpy.DistortionModel.PTLENS:
        a,b,c = calib.terms
        rd1 = partial(ptlens, a=a, b=b, c=c)    
        
    elif calib.model == lensfunpy.DistortionModel.POLY3:
        k1,_,_ = calib.terms
        rd1 = partial(poly3, k1=k1)
        
    elif calib.model == lensfunpy.DistortionModel.POLY5:
        k1,k2,_ = calib.terms
        rd1 = partial(poly5, k1=k1, k2=k2)
        
    else:
        raise NotImplementedError
    
    model, params = _formatDistortionParams(calib)
    
    # calculate the sensor's (half) height, as it is used for scaling within lensfun
    # TODO doesn't take into account the aspect ratio yet
    wFX = 36
    hFX = 24
    dFX = np.sqrt(wFX**2 + hFX**2)
    alpha = np.arcsin(wFX/dFX)
    d = dFX*mod.lens.crop_factor
    h = np.cos(alpha)*d
    
    sensorHalfHeight = h/2
    sensorHalfDiagonal = d/2
    
    X = np.linspace(0, sensorHalfDiagonal, 100)
    Xscaled = X/sensorHalfHeight
    
    fig, ax = drawLinePlot(X, 
                           rd1(Xscaled)*sensorHalfHeight,
                           xlabel='$h\;(\mathrm{mm})$',
                           ylabel='$dD/dh\;(\mathrm{mm}^{-1})$',
                           title=model + '(' + params + ')',
                           )
    ax.set_xlim([0, sensorHalfDiagonal])
    # shade y>0 and y<0 with colors, to indicate pincushion vs. barrel distortion
    ymin, ymax = ax.get_ylim()
    ax.autoscale(False)
    polyPincushion = poly_between([0,sensorHalfDiagonal], 0, ymax)
    polyPincushionArtist = ax.fill(*polyPincushion, facecolor='PeachPuff')[0]
    polyBarrel = poly_between([0,sensorHalfDiagonal], ymin, 0)
    polyBarrelArtist = ax.fill(*polyBarrel, facecolor='LightBlue')[0]

    ax.legend([polyPincushionArtist, polyBarrelArtist], 
              ['pincushion', 'barrel'],
              loc='lower right')
    
    return fig, ax

def _formatDistortionParams(calib):
    pround = partial(round, ndigits=7)
    if calib.model == lensfunpy.DistortionModel.PTLENS:
        model = 'ptlens'
        a,b,c = calib.terms
        params = '{}, {}, {}'.format(pround(a), pround(b), pround(c))
        
    elif calib.model == lensfunpy.DistortionModel.POLY3:
        model = 'poly3'
        k1,_,_ = calib.terms
        params = '{}'.format(pround(k1))
        
    elif calib.model == lensfunpy.DistortionModel.POLY5:
        model = 'poly5'
        k1,k2,_ = calib.terms
        params = '{}, {}'.format(pround(k1),pround(k2))
        
    else:
        raise NotImplementedError
    
    return model, params

def drawReferenceStars(mapping, **kw):
    """
    Draws reference stars on top of the original image of an unresampled 
    astrometry-based mapping by using the Tycho-2 catalogue of the Vizier web service.
    
    :param auromat.mapping.astrometry.BaseAstrometryMapping mapping: the mapping
    :rtype: tuple(Figure,Axes)
    
    For other parameters, see :func:`drawIndxPlot`.
    """
    return drawIndxPlot(mapping.rgb_unmasked, wcsPath=mapping.wcsHeader,
                        useWebCatalog=True, **kw)

def drawIndxPlot(imagePath, axyPath=None, xylsPath=None, matchPath=None,
                 wcsPath=None, newWcsPathOrHeader=None, maxAxyObjects=40, mask=None,
                 useWebCatalog=False, webCatalogLimit=1000, xylsSources=None, scale=5,
                 figax=None):
    """
    Draws reference and/or image-extracted stars from astrometry.net's .axy and .xyls files,
    or for reference stars as well using the Vizier webservice with the Tycho-2 catalogue.
    
    The matching star quad as found by astrometry.net and stored in .match files can also be
    visualized.
    
    :param str|array imagePath: the image to use as background, either a path or
                                an RGB/grayscale image array of uint8/uint16 type
    :param str axyPath: path to .axy file containing extracted stars
    :param str xylsPath: path to .xyls file containing reference stars
    :param str matchPath: path to .match file containing the matched quad (see astrometry.net docs)
    :param wcsPath, newWcsPathOrHeader: if given together with xlysPath, 
                    then the xyls pixel positions are recalculated
    :param useWebCatalog: if True and wcsPath is given, then astroquery
                          is used to obtain reference stars
    :param webCatalogLimit: the maximum number of stars to fetch from the web catalog 
    :param float scale: the size of the circles drawn around stars
    :param int maxAxyObjects: the maximum number of extracted stars to draw
    :param array-like xylsSources: an x,y pixel coordinate array to use as reference stars 
    :param figax: an existing (Figure,Axes) tuple to use; if given, the background image is not drawn;
                  see :func:`auromat.draw_helpers.loadFigImage`
    :param mask: boolean array of the size of the image, True=don't draw stars here
    :rtype: tuple(Figure,Axes)
    """
    assert axyPath or xylsPath or xylsSources or useWebCatalog
    if axyPath:
        xSource, ySource, fluxSource = auromat.fits.readXy(axyPath, sort=True, retSortField=True)
        sources = list(zip(xSource, ySource, fluxSource))
    
    if xylsSources:
        refSources = xylsSources
    elif xylsPath:
        if wcsPath:
            assert newWcsPathOrHeader
            xRefSource, yRefSource = auromat.fits.recomputeXylsPixelPositions(xylsPath, wcsPath,
                                                                              newWcsPathOrHeader)
        else:
            xRefSource, yRefSource = auromat.fits.readXy(xylsPath)
        refSources = list(zip(xRefSource, yRefSource))
    elif useWebCatalog:
        assert wcsPath, 'useWebCatalog can only be used together with wcsPath'
        if isinstance(wcsPath, string_types):
            header = auromat.fits.readHeader(wcsPath)
        else:
            header = wcsPath
        xRefSource, yRefSource = auromat.fits.getCatalogStars(header, limit=webCatalogLimit)
        refSources = list(zip(xRefSource, yRefSource))
    else:
        refSources = None
    
    if matchPath:
        quad = auromat.fits.readQuadMatch(matchPath)
    
    if figax:
        fig, ax = figax[0], figax[1]
    else:
        fig, ax = loadFigImage(imagePath)
    
    if mask is not None:
        if axyPath:
            sources = [x_y for x_y in sources if not mask[x_y[1],x_y[0]]]
        if refSources:
            # The reference sources can include stars slightly outside the image bounds.
            # We clip such sources to the image bounds for checking against the mask.
            clipped = np.array(refSources)
            clipped[:,0] = np.clip(clipped[:,0], 0, mask.shape[1]-1)
            clipped[:,1] = np.clip(clipped[:,1], 0, mask.shape[0]-1)
            refSources = [ref_clip[0] for ref_clip in zip(refSources, clipped) 
                          if not mask[ref_clip[1][1], ref_clip[1][0]]]           
    
    # use only the brightest extracted sources    
    # (catalog stars are already nicely distributed and not overcrowded)
    if axyPath and maxAxyObjects:
        sources = sources[:maxAxyObjects]
    
    redRadius = int(round(6*scale))
    greenRadius = int(round(4*scale))
    thickness = int(round(2*scale))
    
    if axyPath:
        sources = np.asarray(sources)
        if len(sources) > 0:
            if refSources:
                # disable alpha to make matches more clear
                alpha = np.ones(len(sources))
            else:
                flux = sources[:,2]
                flux_min, flux_max = np.min(flux), np.max(flux)
                # scale flux to [0.5,1] so that it can be used as alpha
                alpha = 0.7*(flux-flux_min)/(flux_max-flux_min) + 0.3
            colors = np.zeros((len(sources), 4), float)
            colors[:,0] = 1 # red
            colors[:,3] = alpha
            _circles(sources[:,0], sources[:,1], redRadius, ax, edgecolor=colors, lw=thickness, facecolor='')
    
    if refSources:
        refSources = np.asarray(refSources)
        if len(refSources) > 0:
            _circles(refSources[:,0], refSources[:,1], greenRadius, ax, edgecolor='lime', lw=thickness, facecolor='')
    
    if matchPath:
        # close the polygon
        quad = np.vstack((quad, [quad[0]]))
        ax.plot(quad[:,0], quad[:,1], c='lime', lw=max(1, thickness//2))
    
    return fig, ax

def drawConstellations(figax, wcsPath, clipPoly=None, labels=True, lineStyles=('-',':'),
                       lineThickness=5, fontsize=70, colors=None, alpha=0.5, padding=10):
    """ Plot constellation patterns.

    This does not provide extraordinarily high precision; it's just a cosmetic display that sketches out
    the appropriate stick figures on the sky.

    Based on IDL constellations.pro by Marshall Perrin.
    Constellation data is taken from Xephem, courtesy of Elwood Downey. 
    
    See :mod:`auromat.coordinates.constellations` module for licence text.

    :see: https://github.com/mperrin/misc_astro/blob/master/constellations.py
    :param figax: figimage, as produced by :func:`auromat.draw_helpers.loadFigImage`
    :param wcsPath: path to FITS file containing WCS header
    :param clipPoly: (n,2) x,y polygon points; restricts drawing to the region of the polygon
    :param label: whether to draw labels
    :param lineStyles: tuple(standard style, dotted style)
    :param lineThickness:
    :param string|list|dict colors: 
             color of lines and labels, can be:
              - single color name
              - list of color names which will be cycled
              - mapping from constellation name to color name
              The default is a list of 9 colours.
    :param alpha: alpha of lines and labels
    :param padding: spacing at line ends in pixels
    """
    header = readHeader(wcsPath)
    w, h = header['IMAGEW'], header['IMAGEH']
    wcs = WCS(header)
    ax = figax[1]
    labelBorder = w*0.05
    
    if isinstance(colors, dict):
        constellationColor = lambda name: colors[name]
    else:
        if not colors:
            colors = itertools.cycle(['white', 'lime', 'red', 'orange', 'cyan', 'magenta', 'lightblue', 'hotpink', 'yellow'])
        elif isinstance(colors, string_types):
            colors = itertools.cycle([colors])
        else:
            colors = itertools.cycle(colors)
        constellationColor = lambda _: next(colors)
    
    if clipPoly is not None:
        clip_path = Polygon(clipPoly, visible=False)
        # have to add the polygon, otherwise strange things happen
        ax.add_patch(clip_path)
        clip_path_contains = clip_path.get_path().contains_point
    else:
        clip_path = None
        clip_path_contains = None
    
    for name, point_list in constellations.data.items():
        points = np.asarray(point_list)
        
        drawtype = points[:,0]
        ra_degrees = points[:,1] / 1800. * 15 # 15 = hours to degrees
        dec_degrees = points[:,2] / 60.
                
        x, y = wcs.wcs_world2pix(ra_degrees, dec_degrees, 0)
        
        # With the TAN WCS projection there is only half of the celestial plane
        # defined, which means that for some RA/Dec values there is no
        # equivalent pixel value. Still, wcs_world2pix returns bogus coordinates
        # in these cases instead of NaN. To identify these bogus results, we
        # convert x,y back to RA,Dec and see if there is a mismatch.
        # TODO fixed in master, rework once astropy 0.4.2 is released
        #   (see https://github.com/astropy/astropy/pull/2965)
        ra_check, dec_check = wcs.wcs_pix2world(x, y, 0)
        if not np.allclose(ra_degrees, ra_check) or not np.allclose(dec_degrees, dec_check):
            continue
                
        if np.all((x<0) | (x>=w) | (y<0) | (y>=h)):
            continue
        
        if clip_path and not any(clip_path_contains([x_,y_]) for x_,y_ in zip(x,y)):
            # this check is necessary as the label logic expects at least one star
            # inside the clipping area
            continue
                
        color_ = constellationColor(name)
        
        for i in range(0, len(points)):
            if drawtype[i] == 0:    
                continue # don't draw lines, just move for type 0
            
            ls = lineStyles[1] if drawtype[i] == 2 else lineStyles[0]
            
            if padding > 0:
                originalStart = np.array([x[i-1], y[i-1]])
                originalEnd = np.array([x[i], y[i]])
                vec = originalEnd - originalStart
                length = np.sqrt((vec*vec).sum(axis=0))
                direction = vec/length
                newStart = originalStart + direction*padding
                newEnd = originalStart + direction*(length-padding)
                
                ax.plot([newStart[0],newEnd[0]], [newStart[1],newEnd[1]], linestyle=ls,
                        clip_path=clip_path, 
                        lw=lineThickness, color=color_, alpha=alpha)
            else:
                ax.plot(x[i-1:i+1], y[i-1:i+1], linestyle=ls,
                        clip_path=clip_path, 
                        lw=lineThickness, color=color_, alpha=alpha)
            
        if labels:
            name = name.replace('_', ' ')
            xl = np.unique(x)
            yl = np.unique(y)
            labelx = np.clip(np.mean(xl), labelBorder, w-labelBorder)
            labely = np.clip(np.mean(yl), labelBorder, h-labelBorder)
            # check if label is outside clipping area
            if clip_path and not clip_path_contains([labelx,labely]):
                # find all stars that are inside the clip path and use only them for positioning
                xyl = filter(lambda x_y: x_y[0]>=0 and x_y[0]<w and x_y[1]>=0 and x_y[1]<h and 
                                         clip_path_contains([x_y[0],x_y[1]]), zip(x,y))
                xl = np.unique([x_ for x_,_ in xyl])
                yl = np.unique([y_ for _,y_ in xyl])
                labelx = np.clip(np.mean(xl), labelBorder, w-labelBorder)
                labely = np.clip(np.mean(yl), labelBorder, h-labelBorder)
            ax.text(labelx, labely, name, fontsize=fontsize, color=color_, 
                    horizontalalignment='center', verticalalignment='center', alpha=alpha,
                    clip_path=clip_path) 

    return figax

def getFixedConstellationColors(colors=None):
    """
    Tries to determine a color for each constellation such that neighboring
    constellations have different colors.
    
    :param colors: set of base colors to use (optional)
    :return: a mapping from constellation names to color strings
    :rtype: dict
    """
    if colors is None:
        colors = {'white', 'lime', 'red', 'orange', 'cyan', 'magenta', 'lightblue', 'hotpink', 'yellow'}
    else:
        colors = set(colors)
        
    def find_neighbors(x, tri):
        """see https://stackoverflow.com/a/17811731/60982"""
        return list(set(indx for simplex in tri.vertices if x in simplex for indx in simplex if indx !=x))
    
    keys = list(constellations.data.keys())
    # we use the "middle" point of each constellation as input for neighbor analysis
    points = np.array([p[len(p)//2][1:] for p in constellations.data.values()])
    tri = Delaunay(points)
    constellationColors = {} # mapping from constellation index to color name
    for i in range(len(points)):
        # determine colors that are already used
        indices = find_neighbors(i, tri) + [i]
        indicesDone = list(filter(lambda i: i in constellationColors, indices))
        colorsDone = map(lambda i: constellationColors[i], indicesDone)
        remainingColors = colors - set(colorsDone)
        remainingIndices = set(indices) - set(indicesDone)
        for i, color in zip(remainingIndices, remainingColors):
            constellationColors[i] = color
    
    constellationColors = {keys[i]: c for i,c in constellationColors.items()}
    return constellationColors 

def drawParallelsAndMeridians(mapping, boundingBox=None, labelLon=None, labelLat=None, 
                              lineThickness=5, color='white', alpha=0.5, figax=None):
    """
    Draws parallels and meridians on top of an unresampled mapping.
    
    :param auromat.mapping.mapping.BaseMapping mapping: the unresampled mapping
    :param boundingBox: the bounding box for which to draw parallels/meridians (optional)                        
    :param labelLon: the longitude along which the latitude labels are drawn (optional)
    :param labelLat: the latitude along which the longitude labels are drawn (optional) 
    :param lineThickness: line width
    :param color: color of the parallels and meridians (matplotlib color spec)
    :param alpha: alpha value of the parallels and meridians
    :param figax: figimage, as produced by :func:`auromat.draw_helpers.loadFigImage`
    """
    # The idea is to resample the pixel coordinates into a regular grid such
    # that we have a lookup table from lat/lon to pixel position.

    if boundingBox is None:
        boundingBox = mapping.boundingBox
    
    # First, we interpolate to a fine-grained resolution.
    # When drawing, we choose certain integer parallels and meridians out of that.
    # The means that most of the interpolated values are not actually used.
    interpolateEvery = 0.2 # degrees
    pxPerDeg = 1/interpolateEvery
    assert modf(pxPerDeg)[0] == 0, '1 mod interpolateEvery = 0 is not satisfied'
    
    h,w = mapping.latsCenter.shape[0:2]
    # data is x,y pixel coordinates, (h,w,2)
    data = np.dstack(np.meshgrid(np.arange(w), np.arange(h))).astype(np.float)
        
    _,_,latGrid,lonGrid,pxcoords = _resample(mapping.latsCenter.filled(np.nan), 
                                             mapping.lonsCenter.filled(np.nan), mapping.altitude,
                                             data, lambda: mapping.outline, boundingBox,
                                             (pxPerDeg,pxPerDeg), boundingBox.containsDiscontinuity, 
                                             boundingBox.containsPole, method='mean')
    
    assert pxcoords.ndim == 3 and pxcoords.shape[2] == 2
    
    drawEvery = 2 # degrees, must be an integer
    
    if figax:
        fig, ax = figax[0], figax[1]
    else:
        fig, ax = loadFigImage(mapping.rgb_unmasked)
        
    # draw parallels
    # we want to draw integer degrees only, therefore we first identify these
    isLatInteger = np.abs(latGrid[:,0]-np.round(latGrid[:,0]).astype(int)) < 1e-10        
    latIndices = isLatInteger.nonzero()[0]
    if len(latIndices) > 0:
        # retain only those latitudes that are on fixed grid, e.g. -180,-175,-170 etc for drawEvery=5
        gridCheck = np.array([abs(latGrid[i,0]/drawEvery) for i in latIndices])
        latIndices = latIndices[np.abs(gridCheck-np.round(gridCheck).astype(int)) < 1e-10]
    
    for i in latIndices:
        ax.plot(pxcoords[i,:,0], pxcoords[i,:,1], 
                linestyle='--', lw=lineThickness, color=color, alpha=alpha)

    # draw meridians
    isLonInteger = np.abs(lonGrid[0,:]-np.round(lonGrid[0,:]).astype(int)) < 1e-10
    lonIndices = isLonInteger.nonzero()[0]
    if len(lonIndices) > 0:
        # retain only those longitudes that are on fixed grid, e.g. -90,-85,-80 etc for drawEvery=5
        gridCheck = np.array([abs(lonGrid[0,i]/drawEvery) for i in lonIndices])
        lonIndices = lonIndices[np.abs(gridCheck-np.round(gridCheck).astype(int)) < 1e-10]
    
    for i in lonIndices:
        ax.plot(pxcoords[:,i,0], pxcoords[:,i,1], 
                linestyle='--', lw=lineThickness, color=color, alpha=alpha)
    
    # draw labels
    # we find the parallel and meridian that is closest to the mapping centroid
    # and use that for labelling
    # to get a point which is more in the center, we first mask the far away areas
    if labelLat is None or labelLon is None:
        mapping = mapping.maskedByElevation(10)
    if labelLat is None:
        labelLat = mapping.centroid.lat
    if labelLon is None:
        labelLon = mapping.centroid.lon
    
    axtext = partial(ax.text, fontsize=70, color=color, alpha=alpha,
                     horizontalalignment='center', verticalalignment='center')

    # the longitude where latitude labels are drawn
    # we draw labels between two longitudes
    if len(lonIndices) > 0:
        if boundingBox.containsDiscontinuity:
            angle = 180
            lonsRot = Angle((lonGrid[0,lonIndices] + angle) * u.deg).wrap_at(angle * u.deg).degree
            labelLonRot = Angle((labelLon + angle) * u.deg).wrap_at(angle * u.deg).degree
            distances = lonsRot-labelLonRot
        else:
            distances = lonGrid[0,lonIndices]-labelLon
        labelLonLeftIdx = np.argmin(np.abs(distances))
        labelLonRightIdx = labelLonLeftIdx + 1 if labelLonLeftIdx + 1 < len(lonIndices) else labelLonLeftIdx
        labelLonGridMiddleIdx = (lonIndices[labelLonLeftIdx] + lonIndices[labelLonRightIdx])//2
        
        for i in filter(lambda i: not np.isnan(pxcoords[i,labelLonGridMiddleIdx,0]), latIndices):
            lat = latGrid[i,0]
            if lat>0:
                direction = 'N'
            elif lat<0:
                direction = 'S'
            else:
                direction = ''
            latStr = u'%i\N{DEGREE SIGN}%s' % (np.round(abs(lat)), direction)
            axtext(pxcoords[i,labelLonGridMiddleIdx,0], pxcoords[i,labelLonGridMiddleIdx,1], latStr)
    
    # the latitude where longitude labels are drawn
    if len(latIndices) > 0:
        labelLatLeftIdx = np.argmin(np.abs(latGrid[latIndices,0]-labelLat))
        labelLatRightIdx = labelLatLeftIdx + 1 if labelLatLeftIdx + 1 < len(latIndices) else labelLatLeftIdx
        labelLatGridMiddleIdx = (latIndices[labelLatLeftIdx] + latIndices[labelLatRightIdx])//2
        
        for i in filter(lambda i: not np.isnan(pxcoords[labelLatGridMiddleIdx,i,0]), lonIndices):
            lon = lonGrid[0,i]
            if lon>0:
                direction = 'E'
            elif lon<0:
                direction = 'W'
            else:
                direction = ''
            lonStr = u'%i\N{DEGREE SIGN}%s' % (np.round(abs(lon)), direction)
            axtext(pxcoords[labelLatGridMiddleIdx,i,0], pxcoords[labelLatGridMiddleIdx,i,1], lonStr)
    
    return fig, ax      

def drawDate(figax, mapping, color='white'):
    """
    Writes the mapping date in the top center of the image.
    
    :param figax: figimage, as produced by :func:`draw_helpers.loadFigImage`
    """
    ax = figax[1]
    fontsize = ax.get_xlim()[1] * 0.016
    ax.text(0.5, 0.98, mapping.photoTime.strftime(DatePlotTitleFormat),
            fontsize=fontsize, color=color,
            horizontalalignment='center', verticalalignment='top',
            transform=ax.transAxes)

def _readCorrs(corrPaths):
    """
    
    :param str|iterable corrPaths: paths of .corr files written by astrometry.net
    """
    if isinstance(corrPaths, str):
        corrPaths = [corrPaths]
        
    xFieldArr, yFieldArr = [], []
    xIndexArr, yIndexArr = [], [] 
    for corrPath in corrPaths:
        xField, yField, xIndex, yIndex = auromat.fits.readCorr(corrPath)
        xFieldArr.append(xField)
        yFieldArr.append(yField)
        xIndexArr.append(xIndex)
        yIndexArr.append(yIndex)
    
    return xFieldArr, yFieldArr, xIndexArr, yIndexArr

def drawCorrPlot(corrPaths,
                 title=r'Distances between corresponding stars: $\mu={:0.2f}$, $\sigma={:0.2f}$'):
    """
    Draws a histogram of distances between corresponding stars from an astrometry.net .corr file.
    
    :param corrPaths: path to .corr file or list of paths (histograms are merged)
    :rtype: tuple(Figure,Axes)
    """
    xFieldArr, yFieldArr, xIndexArr, yIndexArr = _readCorrs(corrPaths)
    
    xField = np.concatenate(xFieldArr)
    yField = np.concatenate(yFieldArr)
    xIndex = np.concatenate(xIndexArr)
    yIndex = np.concatenate(yIndexArr)
    
    field = np.transpose([xField, yField])
    index = np.transpose([xIndex, yIndex])
    dist = auromat.utils.vectorLengths(field - index)
    mean, std = np.mean(dist), np.std(dist)
    
    colors = _getMplColors()
    fig,ax = plt.subplots()
    _setMplColors(fig, ax, colors)
    ax.set_xlabel('Distance (pixels)')
    ax.set_ylabel('Count')
    ax.set_title(title.format(mean, std))
    ax.hist(dist, bins=100)
    
    return fig, ax
    
def drawCorrSeqPlot(corrPaths, x=None, xlabel='Frame', retline=False):
    """
    Draws the mean distance of corresponding stars from astrometry.net's .corr files
    for each given file.
    
    :param corrPaths: list of .corr paths
    :param x: x axis values
    :rtype: tuple(Figure,Axes[,Line])
    """
    if x is None:
        x = list(range(len(corrPaths)))
    assert len(x) == len(corrPaths)
    xFieldArr, yFieldArr, xIndexArr, yIndexArr = _readCorrs(corrPaths)
    
    means = []
    stds = []
    for xField, yField, xIndex, yIndex in zip(xFieldArr, yFieldArr, xIndexArr, yIndexArr):    
        field = np.transpose([xField, yField])
        index = np.transpose([xIndex, yIndex])
        dist = auromat.utils.vectorLengths(field - index)
        means.append(np.mean(dist))
        stds.append(np.std(dist))
        
    fig,ax = plt.subplots()
    ax.set_xlabel(xlabel)
    ax.set_ylabel('Mean distance (pixels)')
    ax.set_title(r'Distances between corresponding stars')
    plotline, _, _ = ax.errorbar(x, means, stds, linestyle='None', marker='s')
    ax.set_xlim(x[0]-1, x[-1]+1)
    
    if retline:
        return fig, ax, plotline
    else:
        return fig, ax

def _getHeaders(wcsPathsOrHeadersOrMappings):
    wcsPathsOrHeadersOrMappings = list(wcsPathsOrHeadersOrMappings)
    if not wcsPathsOrHeadersOrMappings:
        return []
    if isinstance(wcsPathsOrHeadersOrMappings[0], string_types):
        headers = list(map(auromat.fits.readHeader, wcsPathsOrHeadersOrMappings))
    elif hasattr(wcsPathsOrHeadersOrMappings[0], 'wcsHeader'):
        headers = [m.wcsHeader for m in wcsPathsOrHeadersOrMappings]
    else:
        headers = wcsPathsOrHeadersOrMappings
    return headers

def drawAstrometryPixelScales(wcsHeaders, x=None, 
                              xlabel='Frame - 1st Frame', ylabel='Pixel scale (arcsec/px)',
                              title='Celestial Sphere Pixel Scale ($\sqrt{CD_{11}^2 + CD_{21}^2}$)',
                              retscatter=False, **kw):
    """
    Draws a scatter plot of the celestial pixel scales of the given astrometric solutions.
    
    :param wcsHeaders: iterable of either WCS header objects or paths to .wcs files or astrometry-based mappings
    :param x: values to use for the x axis; if not given, then x simply increase from 0 to len(wcsHeaders)
    :param bool retscatter: whether to return the `PathCollection` of scatter points
    :rtype: tuple(Figure,Axes[,PathCollection])
    """
    wcsHeaders = list(wcsHeaders)
    if x is None:
        x = list(range(len(wcsHeaders)))
    assert len(x) == len(wcsHeaders)
    
    wcsHeaders = _getHeaders(wcsHeaders)
    pixelScales = list(map(auromat.fits.getPixelScale, wcsHeaders))
    arcsec = (pixelScales*u.deg).to(u.arcsec).value
    
    fig, ax, scatter = _drawScatterPlot(x, arcsec,
                                        xlabel=xlabel, 
                                        ylabel=ylabel, 
                                        title=title,
                                        s=10, retscatter=True, **kw)

    mean = np.mean(arcsec)
    median = np.median(arcsec)
    stddev = np.std(arcsec)
        
    # axhline is not supported by mpld3 yet
    # but we can emulate it with hlines
    #meanline = ax.axhline(mean, color='red')
    #medianline = ax.axhline(median, color='blue')
    meanline = ax.hlines([mean], 0, x[-1]*2, color='red')
    medianline = ax.hlines([median], 0, x[-1]*2, color='blue')
    
    ax.legend([meanline, medianline], 
              ['mean, $\sigma={:0.4f}$'.format(stddev), 'median'],
              loc='upper right', frameon=False)
    
    ax.set_xlim(0, len(wcsHeaders))
    
    if retscatter:
        return fig, ax, scatter
    else:
        return fig, ax
    
def drawAstrometryRotationAngles(wcsHeaders, x=None, 
                                 xlabel='Time', ylabel='Rotation angle (deg)',
                                 title=r'Rotation Angle ($\operatorname{atan}(CD_{21},CD_{11})$)',
                                 retline=False, **kw):
    """
    Draws a line plot of the celestial rotation angles of the given astrometric solutions.
    
    :param wcsHeaders: iterable of either WCS header objects or paths to .wcs files or astrometry-based mappings
    :param x: values to use for the x axis; if not given, then x will be the time axis
    :param bool retline: whether to return the matplotlib `Line` object
    :param kw: additional keywords are handed over to :func:`drawLinePlot`
    :rtype: tuple(Figure,Axes[,Line])
    """
    wcsHeaders = _getHeaders(wcsHeaders)
    
    if x is None:
        x = list(map(auromat.fits.getPhotoTime, wcsHeaders))
    assert len(x) == len(wcsHeaders)
    
    angles = np.asarray(list(map(auromat.fits.getRotationAngle, wcsHeaders)))
    containsDiscontinuity = np.max(angles) - np.min(angles) > 100
    if containsDiscontinuity:
        angles = Angle((angles + 180) * u.deg).wrap_at(180 * u.deg).degree
    
    fig, ax, line = drawLinePlot(x, angles,
                                 xlabel=xlabel, ylabel=ylabel, title=title,
                                 retline=True, **kw)
    
    if containsDiscontinuity:
        def fixAxis(x,pos):
            deg = Angle((x + 180) * u.deg).wrap_at(180 * u.deg).degree
            s = '{:g}'.format(deg) # removes trailing .0
            return s.replace('-', u'\u2212') # uses unicode minus (longer)
            # Note: this should rather make use of ScalarFormatter but I couldn't figure it out
        ax.yaxis.set_major_formatter(FuncFormatter(fixAxis))
    
    if retline:
        return fig, ax, line
    else:
        return fig, ax
    
def drawCD11CD21(wcsHeaders, xlabel='$CD_{11}$', ylabel='$CD_{21}$',
                 title='WCS Transformation Matrix Values',
                 retline=False, **kw):
    """
    Draws a line plot of the CD11 and CD21 values of the given astrometric solutions
    together with a reference circle derived from a fixed (median) celestial pixel scale.
    
    :param wcsHeaders: iterable of either WCS header objects or paths to .wcs files or astrometry-based mappings
    :param bool retline: whether to return the matplotlib `Line` object
    :param kw: additional keywords are handed over to :func:`drawLinePlot`
    :rtype: tuple(Figure,Axes[,Line])
    """
    wcsHeaders = _getHeaders(wcsHeaders)
    cd11 = [h['CD1_1'] for h in wcsHeaders]
    cd21 = [h['CD2_1'] for h in wcsHeaders]
    
    pixelScales = list(map(auromat.fits.getPixelScale, wcsHeaders))
    pixelScaleMedian = np.median(pixelScales)
    
    fig, ax, line = drawLinePlot(cd11, cd21,
                                 xlabel=xlabel, ylabel=ylabel, 
                                 title=title, retline=True, **kw)
    
    circle = plt.Circle((0,0), pixelScaleMedian, fill=False)
    ax.add_patch(circle)
    
    pixelScaleMedianArcSec = (pixelScaleMedian*u.deg).to(u.arcsec).value
    
    ax.legend([circle], 
              ['{0:0.2f}'.format(pixelScaleMedianArcSec) + ' arcsec/px (median)'],
              loc='upper right', frameon=False)
    
    if retline:
        return fig, ax, line
    else:
        return fig, ax
    
def drawRaDec(wcsHeaders,
              xlabel='Right ascension (deg)', ylabel='Declination (deg)',
              title='Equatorial Coordinates of Image Centers', 
              retline=False, **kw):
    """
    Draws a line plot of the right ascension and declination values of the given astrometric solutions.
    
    :note: Assumes that CRPIX1,CRPIX2 are the image center.
    :param wcsHeaders: iterable of either WCS header objects or paths to .wcs files or astrometry-based mappings
    :param bool retline: whether to return the matplotlib `Line` object
    :param kw: additional keywords are handed over to :func:`drawLinePlot`
    :rtype: tuple(Figure,Axes[,Line])
    """
    wcsHeaders = _getHeaders(wcsHeaders)
    ra = [h['CRVAL1'] for h in wcsHeaders]
    dec = [h['CRVAL2'] for h in wcsHeaders]
    
    return drawLinePlot(ra, dec,
                        xlabel=xlabel, ylabel=ylabel, 
                        title=title, retline=retline, **kw)
    
def drawRightAscension(wcsHeaders, x=None,
                       xlabel='Time', ylabel='Right ascension (deg)',
                       title='Right Ascension of Image Centers', retline=False, **kw):
    """
    Draws a line plot of the right ascension values of the given astrometric solutions.
    
    :note: Assumes that CRPIX1,CRPIX2 are the image center.
    :param wcsHeaders: iterable of either WCS header objects or paths to .wcs files or astrometry-based mappings
    :param x: values to use for the x axis; if not given, then x will be the time axis
    :param bool retline: whether to return the matplotlib `Line` object
    :param kw: additional keywords are handed over to :func:`drawLinePlot`
    :rtype: tuple(Figure,Axes[,Line])
    """
    wcsHeaders = _getHeaders(wcsHeaders)
    
    if x is None:
        x = map(auromat.fits.getPhotoTime, wcsHeaders)
    assert len(x) == len(wcsHeaders)
    
    ra = [h['CRVAL1'] for h in wcsHeaders]
    return drawLinePlot(x, ra,
                        xlabel=xlabel, ylabel=ylabel, 
                        title=title, retline=True, **kw)
    
def drawDeclination(wcsHeaders, x=None,
                    xlabel='Time', ylabel='Declination (deg)',
                    title='Declination of Image Centers', retline=False, **kw):
    """
    Draws a line plot of the declination values of the given astrometric solutions.
    
    :note: Assumes that CRPIX1,CRPIX2 are the image center.
    :param wcsHeaders: iterable of either WCS header objects or paths to .wcs files or astrometry-based mappings
    :param x: values to use for the x axis; if not given, then x will be the time axis
    :param bool retline: whether to return the matplotlib `Line` object
    :param kw: additional keywords are handed over to :func:`drawLinePlot`
    :rtype: tuple(Figure,Axes[,Line])
    """  
    wcsHeaders = _getHeaders(wcsHeaders)
    
    if x is None:
        x = map(auromat.fits.getPhotoTime, wcsHeaders)
    assert len(x) == len(wcsHeaders)
    
    dec = [h['CRVAL2'] for h in wcsHeaders]
    
    return drawLinePlot(x, dec,
                        xlabel=xlabel, ylabel=ylabel, 
                        title=title, retline=True, **kw)

def drawCameraFootpoints(mappings,
                         xlabel='Longitude (deg)', ylabel='Latitude (deg)',
                         title='Camera Footpoints', 
                         retline=False, **kw):
    """
    Draws a line plot of the camera footpoints of the given mappings.
    
    :param mapping: iterable of mappings
    :param bool retline: whether to return the matplotlib `Line` object
    :param kw: additional keywords are handed over to :func:`drawLinePlot`
    :rtype: tuple(Figure,Axes[,Line])
    """
    footpoints = [m.cameraFootpoint for m in mappings]
    lat = [f.lat for f in footpoints]
    lon = [f.lon for f in footpoints]
    
    return drawLinePlot(lon, lat,
                        xlabel=xlabel, ylabel=ylabel, 
                        title=title, retline=retline, **kw)

def setColors(figax, bgcolor='white', transparent=False):
    """
    Changes colors of all plot elements to a given color scheme.
    
    :param figax: `Figure` object, or array-like with `Figure`,`Axes` as first two elements 
    :param str bgcolor: background color of the plot, 'white' or 'black'
    :param bool transparent: whether the background outside the plot bounds should be transparent
    """
    try:
        fig, ax = figax[0], figax[1]
    except:
        fig = figax
        ax = fig.gca()
    _setMplColors(fig, ax, _getMplColors(bgcolor, transparent))

def saveFig(outputFile, figax, bgcolor='white', transparent=False, widthPx=None, heightPx=None, dpi=None, format_=None,
            dontClose=False):
    """
    Saves a matplotlib `Figure` to a path on disk or a `File`-like object.
    
    :note: Only one of `widthPx`, `heightPx`, and `dpi` must be defined. If none is given, then default
           values from matplotlib are used.
    :param outputFile: path to file, or `File`-like object
    :param figax: `Figure` object, or array-like with `Figure`,`Axes` as first two elements 
    :param str bgcolor: background color of the plot, 'white' or 'black'
    :param bool transparent: whether the background outside the plot bounds should be transparent
    :param int widthPx: width in pixels
    :param int heightPx: height in pixels
    :param number dpi: dpi in inches
    :param str format_: file format, e.g. 'jpg', 'png', 'svg'; will be derived from extension if `outputFile`
                        is a path name
    :param bool dontClose: whether to close the `Figure` or leave it open for further processing;
                           default is to close it; use `plt.close(fig)` for manual closing
    """
    try:
        fig, ax = figax[0], figax[1]
    except:
        fig = figax
        ax = fig.gca()
    setColors((fig,ax), bgcolor, transparent)
    return _saveFig(outputFile, fig,
                    widthPx=widthPx, heightPx=heightPx, dpi=dpi,
                    format_=format_, dontClose=dontClose)
    
if __name__ == '__main__':
    fig, ax = plt.subplots()
    colors = getFixedConstellationColors({'lime', 'red', 'orange', 'cyan', 'magenta', 'lightblue', 'hotpink', 'yellow'})
    for name, points in constellations.data.items():
        points = np.asarray(points)
        drawtype = points[:,0]
        ra_degrees = points[:,1] / 1800. * 15 # 15 = hours to degrees
        dec_degrees = points[:,2] / 60.
        for i in range(0, len(points)):
            if drawtype[i] == 0:    
                continue
            ax.plot(ra_degrees[i-1:i+1], dec_degrees[i-1:i+1], linestyle='-',
                    color=colors[name])
    
    saveFig('test.svg',(fig,ax))
        
    print(getFixedConstellationColors())


# Copyright European Space Agency, 2013

from __future__ import division, absolute_import

import unittest
from auromat.mapping.mapping import BoundingBox
from numpy.testing.utils import assert_array_almost_equal, assert_array_equal
    

class Test(unittest.TestCase):
    
    def testBB(self):
        bb = BoundingBox(latSouth=-60, lonWest=80, 
                         latNorth=-30, lonEast=85)        
        assert_array_almost_equal(bb.center, [-45.03119418083877, 82.5])
        assert_array_almost_equal(bb.size, [482.39311013217343, 3336.5953086140203])

    def testBBDiscontinuity(self):
        bb = BoundingBox(latSouth=-60.646114098, lonWest=82.7852215499, 
                         latNorth=-38.7515567117, lonEast=-178.546517062)
        assert_array_almost_equal(bb.center, [-54.33647117488648, 132.11935224395])
        assert_array_almost_equal(bb.size, [8084.704893634039, 3464.8889697347718])

    def testBBPole(self):
        # north pole
        bb = BoundingBox(latSouth=60, lonWest=-180, latNorth=90, lonEast=180)
        assert_array_almost_equal(bb.center, [90, 0])
        assert_array_almost_equal(bb.size, [6695.78581964, 6695.78581964])
        
        # south pole
        bb = BoundingBox(latSouth=-90, lonWest=-180, latNorth=-60, lonEast=180)
        assert_array_almost_equal(bb.center, [-90, 0])
        assert_array_almost_equal(bb.size, [6695.78581964, 6695.78581964])
    
    def testBBPoint(self):
        bb = BoundingBox(latSouth=50, lonWest=80, 
                         latNorth=50, lonEast=80)
        assert_array_equal(bb.center, [50,80])
        assert_array_equal(bb.size, 0)
        
    def testBBMerge(self):
        bb1 = BoundingBox(latSouth=-55, lonWest=95, latNorth=-45, lonEast=109)
        bb2 = BoundingBox(latSouth=44, lonWest=-164, latNorth=74, lonEast=-35)
        bb = BoundingBox.mergedBoundingBoxes([bb1,bb2])
        assert_array_equal([bb.latSouth,bb.latNorth,bb.lonWest,bb.lonEast],
                           [bb1.latSouth,bb2.latNorth,bb1.lonWest,bb2.lonEast])
        assert_array_almost_equal(bb.center, [21.136113246, -150])
        # size cannot be checked as the bounding box spans more than 180deg longitude
        # which is not supported currently
